import os
import sys
import yaml
import time
import difflib
from collections import defaultdict
from typing import List, Dict, Any

import rich_click as click
from prettytable import PrettyTable, TableStyle
from rich.style import Style
from rich.text import Text
from rich.pretty import pprint
from rich.console import Console, ConsoleOptions
from rich.panel import Panel
from rich.syntax import Syntax
from pycomex.functional.experiment import Experiment
from pycomex.utils import dynamic_import

from chem_mat_data.config import Config
from chem_mat_data.web import NextcloudFileShare
from chem_mat_data.utils import get_version
from chem_mat_data.utils import PATH
from chem_mat_data.utils import DOCS_PATH
from chem_mat_data.utils import METADATA_PATH
from chem_mat_data.utils import TEMPLATE_ENV
from chem_mat_data.utils import CsvListType
from chem_mat_data.utils import open_file_in_editor
from chem_mat_data.utils import RichMixin

# The path to the "scripts" folder of the experiment modules
SCRIPTS_PATH: str = os.path.join(PATH, 'scripts')
# The path to the "results" folder of the experiment module executions.
RESULTS_PATH: str = os.path.join(PATH, 'scripts', 'results')


class RichDiffDisplay(RichMixin):
    """
    Rich display element that shows the diff between two metadata.yml files.
    """
    def __init__(self, local_file: str, remote_file: str, diff_lines: List[str], changed_lines: int):
        self.local_file = local_file
        self.remote_file = remote_file
        self.diff_lines = diff_lines
        self.changed_lines = changed_lines
        
    def __rich_console__(self, console: Console, options: ConsoleOptions) -> Any:
        
        yield ''
        
        # Show file paths being compared
        from rich.table import Table
        comparison_table = Table(box=None)
        comparison_table.add_column("File", style="magenta", no_wrap=True)
        comparison_table.add_column("Path", style="cyan", no_wrap=False)
        comparison_table.add_row("Local", self.local_file)
        comparison_table.add_row("Remote", "metadata.yml (from server)")
        
        yield comparison_table
        
        # Show the actual diff if there are changes
        if self.changed_lines > 0 and self.diff_lines:
            yield ""
            
            # Limit diff output to first 50 lines to avoid overwhelming display
            display_lines = self.diff_lines[:50]
            if len(self.diff_lines) > 50:
                display_lines.append(f"... ({len(self.diff_lines) - 50} more lines truncated)")
            
            # Create a syntax-highlighted diff
            diff_text = '\n'.join(display_lines)
            syntax = Syntax(diff_text, "diff", theme="monokai", line_numbers=False)
            
            panel = Panel(
                syntax,
                title=f'File Differences (showing first {min(50, len(self.diff_lines))} lines)',
                expand=True
            )
            
            yield panel
            
        # Show diff summary
        yield ""
        if self.changed_lines == 0:
            yield Text(f"✅ Files are identical", style="green")
        else:
            yield Text(f"📊 {self.changed_lines} lines differ between local and remote files", style="yellow")


def is_experiment_archive(folder_path: str) -> bool:
    """
    Helper function to determine if a given ``folder_path`` represents a valid experiment archive.
    """
    if not os.path.isdir(folder_path):
        return False

    data_path = os.path.join(folder_path, 'experiment_data.json')
    meta_path = os.path.join(folder_path, 'experiment_meta.json')
    code_path = os.path.join(folder_path, 'experiment_code.py')
    metadata_path = os.path.join(folder_path, 'metadata.yml')
    if not os.path.exists(data_path) or not os.path.exists(meta_path) or not os.path.exists(code_path) or not os.path.exists(metadata_path):
        return False
    
    return True


class CLI(click.RichGroup):

    # This list contains the string identifiers for the possible strategies that can be used for 
    # the metadata collect method, which will collect the individual metadata files from the 
    # experiment results.
    EXPERIMENT_COLLECT_STRATEGIES: List[str] = ['recent']

    def __init__(self,
                 **kwargs):
        super(CLI, self).__init__(
            invoke_without_command=True,
            **kwargs
        )

        # This config file is a global singleton instance which allows access to the most 
        # important config parameter of the project such as the URL address of the fileshare 
        # server from which the datasets will be downloaded.
        self.config = Config()
        self.file_share = NextcloudFileShare(
            self.config.get_fileshare_url(), 
            **self.config.get_fileshare_parameters('nextcloud')
        )
        self.cache = self.config.cache
        
        self.help = (
            'This is the internal management command line interface for the ChemMatData package. '
            'This CLI will provide certain commands that are useful for maintaining and developing of the '
            'package and the database itself.'
        )
        
        ## -- Adding commands --
        
        # metadata command group
        self.add_command(self.metadata_group)
        self.metadata_group.add_command(self.metadata_collect_command)
        self.metadata_group.add_command(self.metadata_upload_command)
        self.metadata_group.add_command(self.metadata_edit_command)
        self.metadata_group.add_command(self.metadata_diff_command)
        self.metadata_group.add_command(self.metadata_remove_command)
        
        # dataset command group
        self.add_command(self.dataset_group)
        self.dataset_group.add_command(self.dataset_create_command)
        self.dataset_group.add_command(self.dataset_upload_command)

        # docs command group
        self.add_command(self.docs_group)
        self.docs_group.add_command(self.docs_collect_datasets_command)

    ## == METADATA COMMAND GROUP ==
    # This command group is used to manage the metadata.yml file both locally and on the remote server 

    @click.group('metadata', help='Commands for managing the metadata YML of the remote server.')
    @click.pass_obj
    def metadata_group(self,):
        """
        This command group contains commands that can be used to manage the metadata.yml file on the 
        remote file server. This metadata file contains all the necessary information about the 
        datasets and for example determines what information will be shown when using the
        ``list`` command.
        """
        pass
    
    def collect_dataset_archives_map(self, results_path: str) -> Dict[str, List[Experiment]]:
        """
        A helper function which creates a ``dataset_archives_map`` data structure if given the absolute 
        string ``results_path`` of where to search the experiment archives. Returns a dict data structure
        whose string keys are the dataset names and the values are the corresponding Experiment instances
        imported from the experiment archives.
        
        :param results_path: The absolute path to the folder where the experiment archives are stored.
        
        :return: A dict data structure mapping dataset names to Experiment instances.
        """
        click.secho('collecting experiment archives...', fg='bright_black')
        
        # ~ collecting experiment archives
        # We first want to generally collect all the experiment archives here for each dataset.
        experiments: List[Experiment] = []
        for root, folders, files in os.walk(results_path):
            
            for folder in folders:
                
                folder_path = os.path.join(root, folder)
                if is_experiment_archive(folder_path):
                    try:
                        experiment = Experiment.load(folder_path)
                    except Exception:
                        continue
                    
                    # for now we just add the loaded experiment to the list and will then later 
                    # sort out the metadata from it.
                    experiments.append(experiment)
                    
                    experiment_name = experiment.metadata['name']
                    experiment_id = os.path.basename(folder_path)
                    click.secho(f' * collecting experiment {experiment_name}/{experiment_id}', fg='bright_black')
        
        click.secho(f'collected {len(experiments)} experiments')
        
        # Now that we have all the experiments we can sort the experiments into lists of which dataset they belong to.
        # best case there is inly one dataest per experiment, but there are good chances that this will not be the 
        # case. And if there are multiple datasets per experiment, we will have to decide which one to use to get 
        # the metadata from based on the strategy.
        dataset_archives_map: Dict[str, List[Experiment]] = defaultdict(list)
        for experiment in experiments:
            dataset = experiment.parameters['DATASET_NAME']
            dataset_archives_map[dataset].append(experiment)
            
        return dataset_archives_map

    @click.command('collect')
    @click.option('--blank', is_flag=True, help=('If this flag is set, the initial metadata.yml file will be created from scratch.'
                                                 'Otherwise, the remote metadata.yml file will be downloaded and used as a base.'))
    @click.option('-p', '--path', help='The path at which to save the new metadata file', 
                  default=METADATA_PATH, show_default=True)
    @click.option('-s', '--source', help='The source path at which to search for dataset metadata.', default=RESULTS_PATH)
    @click.option('--strategy', help='The strategy to use for collecting the metadata.', 
                  type=click.Choice(EXPERIMENT_COLLECT_STRATEGIES), default='recent')
    @click.option('--verbose', is_flag=True, help='If this flag is set, the command will print additional information.')
    @click.option('--datasets', type=CsvListType(), help=(
        'A list of dataset names to collect metadata for. If not provided, all datasets will be collected.'
    ))
    @click.pass_obj
    def metadata_collect_command(self,
                                 blank: bool,
                                 path: str,
                                 source: str,
                                 strategy: str,
                                 verbose: bool,
                                 datasets: List[str],
                                 ) -> None:
        """
        This command collects the metadata of all datasets that are currently stored on the LOCAL system
        and bundles that information into a single metadata.yml file. Will use the current version of the 
        remote file share metadata as a base unless the `--blank` flag is set.
        """
        time_start: float = time.time()
        
        ## --- Creating the Base ---
        # If the blank flag is NOT explicitly set, we are going to try and download the metadata.yml file from the remote file 
        # share server and use that as a base version which is then updated with the local metadata.
        metadata_all: dict = {'datasets': []}
        if not blank:
            click.echo('downloading the metadata.yml file from the remote file share server...')
            metadata_all.update(self.file_share.fetch_metadata())
        
        ## --- Collecting the Metadata ---
        
        # This method will return a dict data structure that maps the dataset names to lists of Experiment instances
        # that have been loaded from the results folder of completed experiment. For each dataset name key the corresponding
        # value list consists of all experiment archives related to that dataset.
        dataset_archives_map: Dict[str, List[Experiment]] = self.collect_dataset_archives_map(source)
        
        if datasets:
            dataset_archives_map = {
                name: experiments
                for name, experiments in dataset_archives_map.items()
                if name in datasets
            }
        
        click.echo(f'updating information on {len(dataset_archives_map)} datasets: {" ".join(dataset_archives_map.keys())}')
        
        # This data structure will hold the metadata for each dataset that we have collected from the experiments.
        dataset_metadata_map: Dict[str, dict] = {}
        for dataset, experiments in dataset_archives_map.items():
            
            if strategy == 'recent':
                experiment = sorted(experiments, key=lambda e: e.metadata['end_time'])[-1]
        
            # The actual experiment metadata that we are interested in will be stored as an additional artifact 
            # called "metadata.yml" in the experiment folder. We will load this file and then add the metadata
            # to the dataset_metadata_map.
            metadata_path: str = os.path.join(experiment.path, 'metadata.yml')
            with open(metadata_path, mode='r') as file:
                metadata: dict = yaml.load(file, Loader=yaml.FullLoader)
            
            dataset_metadata_map[dataset] = metadata

        # --- merging the metadata ---
        # The structure of the overall metadata is such that there is an additional key for each dataset and the 
        # corresponding value is again a dict-like object that contains the metadata for that dataset.
        click.echo('merging the metadata...')
        for dataset, metadata in dataset_metadata_map.items():
            metadata_all['datasets'][dataset] = metadata
        
        # --- saving the file ---
        # In the end we can save the collected metadata into a file and place it at the specified path.
        path = os.path.expanduser(path)
        click.echo(f'saving the metadata file @ {path}...')
        with open(path, 'w') as file:
            yaml.dump(metadata_all, file, sort_keys=True, indent=4)
            
        time_end: float = time.time()
        
        click.secho(f'✅ collected metadata.yml file in {time_end - time_start:.2f} seconds', fg='green')
            
        if verbose:
            pprint(metadata_all, max_string=100)
            
        click.secho()
            
    @click.command('upload')
    @click.option('-p', '--path', help='The path to the metadata file that should be uploaded.', 
                  default=METADATA_PATH, show_default=True)
    @click.pass_obj
    def metadata_upload_command(self,
                                path: str
                                ) -> None:
        """
        Uploads a local metadata.yml file to the remore file share server. Note that this will overwrite the 
        version of the metadata file that is currently stored on the server.
        """
        if not os.path.exists(path):
            click.secho(f'Error: The specified metadata file "{path}" does not exist.', fg='red')
            sys.exit(1)

        click.secho('uploading the metadata file to the remote file share server...')        
        self.file_share.upload('metadata.yml', path)
        click.secho('✅ uploaded metadata.yml', fg='green')
        click.secho()
        
    @click.command('edit')
    @click.option('-p', '--path', help='The path to the metadata file that should be uploaded.', default='./metadata.yml')
    @click.pass_obj
    def metadata_edit_command(self,
                              path: str,
                              ) -> None:
        """
        Opens the metadata.yml file in the default editor of the system.
        """
        if not os.path.exists(path):
            click.secho(f'Error: The specified metadata file "{path}" does not exist.', fg='red')
            sys.exit(1)
        
        open_file_in_editor(path)

    @click.command('diff', short_help='Compare local metadata.yml with remote version')
    @click.option('--local-file', '-f', type=click.Path(exists=True, dir_okay=False, readable=True), 
                  default=METADATA_PATH, show_default=True,
                  help='Path to the local metadata.yml file to compare')
    @click.pass_obj
    def metadata_diff_command(self, local_file: str) -> None:
        """
        Compares a local metadata.yml file with the remote metadata.yml file from the server.
        Shows the number of different lines and displays a visual diff of the changes.
        
        By default, uses the package metadata.yml file for comparison.
        """
        
        ## --- Fetch Remote Metadata ---
        try:
            # Try to fetch the remote metadata
            self.file_share.fetch_metadata(force=True)
            
            if not hasattr(self.file_share, 'metadata') or self.file_share.metadata is None:
                click.secho('❌ Could not fetch remote metadata.yml file', fg='red')
                sys.exit(1)
                
        except Exception as e:
            click.secho(f'❌ Error fetching remote metadata: {str(e)}', fg='red')
            sys.exit(1)
        
        ## --- Read Local File ---
        try:
            with open(local_file, 'r', encoding='utf-8') as f:
                local_content = f.read()
        except Exception as e:
            click.secho(f'❌ Error reading local file: {str(e)}', fg='red')
            sys.exit(1)
        
        ## --- Normalize Both Files for Comparison ---
        try:
            # Parse local file to normalize it
            local_data = yaml.safe_load(local_content)
            # Convert both to the same YAML format for fair comparison
            local_normalized = yaml.dump(local_data, default_flow_style=False, sort_keys=True, indent=2)
            remote_normalized = yaml.dump(self.file_share.metadata, default_flow_style=False, sort_keys=True, indent=2)
        except Exception as e:
            click.secho(f'❌ Error normalizing YAML data: {str(e)}', fg='red')
            sys.exit(1)
        
        ## --- Calculate Diff ---
        local_lines = local_normalized.splitlines(keepends=False)
        remote_lines = remote_normalized.splitlines(keepends=False)
        
        # Generate unified diff
        diff_lines = list(difflib.unified_diff(
            local_lines,
            remote_lines,
            fromfile=f'local/{os.path.basename(local_file)}',
            tofile='remote/metadata.yml',
            lineterm=''
        ))
        
        # Count changed lines (lines starting with + or - but not +++ or ---)
        changed_lines = sum(1 for line in diff_lines 
                           if (line.startswith('+') and not line.startswith('+++')) or
                              (line.startswith('-') and not line.startswith('---')))
        
        ## --- Display Results ---
        rich_diff = RichDiffDisplay(
            local_file=local_file,
            remote_file="remote/metadata.yml",
            diff_lines=diff_lines,
            changed_lines=changed_lines
        )
        click.echo(rich_diff)

    @click.command('remove', short_help='Remove a dataset entry from the local metadata.yml file')
    @click.argument('dataset_name', type=str)
    @click.option('--metadata-path', '-f', type=click.Path(exists=True, dir_okay=False, readable=True, writable=True),
                  default=METADATA_PATH, show_default=True,
                  help='Path to the local metadata.yml file to modify')
    @click.option('--backup', is_flag=True, default=True, show_default=True,
                  help='Create a backup copy before modifying the file')
    @click.pass_obj
    def metadata_remove_command(self, dataset_name: str, metadata_path: str, backup: bool) -> None:
        """
        Removes a dataset entry from the local metadata.yml file.

        DATASET_NAME: Name of the dataset to remove from the metadata file
        """

        ## --- Read and Parse Local File ---
        try:
            with open(metadata_path, 'r', encoding='utf-8') as f:
                metadata_content = f.read()
            metadata = yaml.safe_load(metadata_content)
        except Exception as e:
            click.secho(f'❌ Error reading metadata file: {str(e)}', fg='red')
            sys.exit(1)

        ## --- Check if Dataset Exists ---
        if 'datasets' not in metadata:
            click.secho('❌ No datasets section found in metadata file', fg='red')
            sys.exit(1)

        if dataset_name not in metadata['datasets']:
            click.secho(f'❌ Dataset "{dataset_name}" not found in metadata file', fg='red')
            available_datasets = list(metadata['datasets'].keys())
            if available_datasets:
                click.secho(f'Available datasets: {", ".join(sorted(available_datasets))}', fg='yellow')
            sys.exit(1)

        ## --- Create Backup if Requested ---
        if backup:
            backup_path = f"{metadata_path}.backup"
            try:
                import shutil
                shutil.copy2(metadata_path, backup_path)
                click.secho(f'📋 Created backup at: {backup_path}', fg='bright_black')
            except Exception as e:
                click.secho(f'⚠️ Warning: Could not create backup: {str(e)}', fg='yellow')

        ## --- Remove Dataset ---
        removed_dataset = metadata['datasets'].pop(dataset_name)

        ## --- Write Updated Metadata ---
        try:
            with open(metadata_path, 'w', encoding='utf-8') as f:
                yaml.dump(metadata, f, default_flow_style=False, sort_keys=True, indent=4)
            click.secho(f'✅ Successfully removed dataset "{dataset_name}" from {metadata_path}', fg='green')

            # Show some info about what was removed
            compounds = removed_dataset.get('compounds', 'Unknown')
            targets = removed_dataset.get('targets', 'Unknown')
            click.secho(f'   Removed dataset had {compounds} compounds and {targets} targets', fg='bright_black')

        except Exception as e:
            click.secho(f'❌ Error writing updated metadata file: {str(e)}', fg='red')
            sys.exit(1)

    ## == DATASET COMMAND GROUP ==
    # all the commands related to the interaction with the local datasets - such as the (re)creation of
    # datasets or the uploading to the remote file share server.
        
    @click.group('dataset', help=('Commands for managing the datasets on the local system.'))
    @click.pass_obj
    def dataset_group(self,
                      ) -> None:
        pass
    
    def collect_dataset_experiment_map(experiments_path: str) -> Dict[str, Experiment]:
        """
        A helper function which creates a ``dataset_experiment_map`` data structure if given 
        the absolute string ``experiments_path`` of where to search the experiment modules. 
        Returns a dict data structure whose string keys are the dataset names and the values 
        are the corresponding Experiment instances imported from the experiment modules.
        
        :param experiments_path: The absolute path to the folder where the experiment modules are stored.
        
        :return: A dict data structure mapping dataset names to Experiment instances.
        """
        # In this data structure we are going to map the string dataset names to the imported
        # Experiment instances that can potentially be executed to create the corresponding 
        # datasets.
        dataset_experiment_map: Dict[str, Experiment] = {}
        
        file_names: List[str] = os.listdir(experiments_path)
        for file_name in file_names:
            file_path = os.path.join(experiments_path, file_name)
            if os.path.isfile(file_path) and file_name.endswith('.py'):
                
                # Each python module we find, we'll import and check if it has the "DATASET_NAME"
                # attribute. If it does, we'll add it to the list of available experiment modules.
                module = dynamic_import(file_path)
                if hasattr(module, 'DATASET_NAME') and hasattr(module, '__experiment__'):
                    experiment = getattr(module, '__experiment__')
                    dataset_name = experiment.parameters['DATASET_NAME']
                    dataset_experiment_map[dataset_name] = experiment
                    
        return dataset_experiment_map
    
    @click.command('create', short_help='Creates a new dataset using the experiment modules.')
    @click.option('--scripts-path', help='The path to the scripts folder.', default=SCRIPTS_PATH)
    @click.option('--all', help='If this flag is set, all available datasets will be created.', is_flag=True)
    @click.argument('name', type=str)
    @click.pass_obj
    def dataset_create_command(self,
                               scripts_path: str,
                               name: str,
                               all: bool,
                               ) -> None:
        """
        This command can be used to trigger the execution of one or multiple experiment modules which 
        will re-create a dataset.
        
        The NAME argument may either be the string name of a single dataset to be created or a 
        comma separated list of dataset names.
        """
        # ~ collecting experiment modules
        # First of all we construct a data structure that tells us which experiment modules are actually 
        # available to create which datasets.
        click.secho('collecting available experiment modules...', color='bright_black')
        
        # This method creates a dict data structure that maps the dataset names to the Experiment instances
        # that are imported from the experiment modules.
        dataset_experiment_map: Dict[str, Experiment] = self.collect_dataset_experiment_map(scripts_path)
        click.secho(f'found {len(dataset_experiment_map)} experiment modules')

        # ~ parsing the name argument
        # The name argument may be a single dataset name or a comma separated list of dataset names.
        if all:
            dataset_names: List[str] = list(dataset_experiment_map.keys())
        else:
            dataset_names: List[str] = [n.replace(' ', '') for n in name.split(',')]

        # ~ executing the experiments
        # We can fetch the experiment instances from the map that we previously constructed and 
        # we can forcefully execute these experiements by using the "run" method.
        click.secho('running the experiment modules. Beware that this will take some time...')
        for dataset_name in dataset_names:
            click.secho(f'⭐ running "{dataset_name}"', fg='yellow')
            experiment = dataset_experiment_map[dataset_name]
            experiment.run()
        
        click.secho('✅ all experiments have been executed', fg='green')
        
    @click.command('upload', short_help='Uploads local datasets to the remote file share server.')
    @click.argument('name', type=str, required=False)
    @click.option('--results-path', help='The path to the results folder.', default=RESULTS_PATH)
    @click.option('-p', '--path', help='The path to the dataset file that should be uploaded.')
    @click.option('--strategy', help='The strategy to use for selecting the experiment to use for the upload.',
                    type=click.Choice(EXPERIMENT_COLLECT_STRATEGIES), default='recent')
    @click.option('--all', help='If this flag is set, all available datasets will be uploaded.', is_flag=True)
    @click.pass_obj
    def dataset_upload_command(self,
                               name: str,
                               results_path: str,
                               path: str,
                               strategy: str,
                               all: bool,
                               ) -> None:
        """
        This command will upload the graph dataset MPACK file and the raw CSV file for a given dataset
        NAME to the remote file share server. If the `--all` flag is set, all available datasets will be
        uploaded. Note that this will overwrite any existing dataset files on the remote server.

        The NAME argument may either be the string name of a single dataset to be uploaded or a
        comma-separated list of dataset names.
        """
        
        # This method will return a dict data structure that maps the dataset names to lists of Experiment instances
        # that have been loaded from the results folder of completed experiment. For each dataset name key the corresponding
        # value list consists of all experiment archives related to that dataset.
        dataset_archives_map: Dict[str, List[Experiment]] = defaultdict(list)
        dataset_archives_map.update(self.collect_dataset_archives_map(results_path))
        click.secho(f'found {len(dataset_archives_map)} datasets')
            
        # ~ parsing the name argument
        # The name argument may be a single dataset name or a comma separated list of dataset names.
        if all:
            dataset_names: List[str] = list(dataset_archives_map.keys())
        elif name:
            dataset_names: List[str] = [n.replace(' ', '') for n in name.split(',')]
        else:
            click.secho('❌ Error: Must specify either dataset NAME(s) or use --all flag', fg='red')
            sys.exit(1)
            
        # ~ uploading the datasets
        # Now we can go through the dataset archives map and upload the dataset files to the file share server.
        click.secho('uploading the datasets...')
        for dataset_name in dataset_names:
            
            experiments = dataset_archives_map[dataset_name]
            if len(experiments) == 0:
                click.secho(f'❗ no experiment archives found for dataset "{dataset_name}"', fg='red')
                continue
            
            click.secho(f'⭐ dataset "{dataset_name}"', fg='yellow')
            
            # First we need to select a single experiment to use as the basis of the upload for each dataset 
            # based on the selected strategy.
            if strategy == 'recent':
                experiment = sorted(experiments, key=lambda e: e.metadata['end_time'])[-1]
        
            # The actual dataset file that we are interested in will be stored as an additional artifact
            # called "{dataset_name}.xxx" in the experiment folder. We will load this file and then
            # upload it to the file share server.
            
            mpack_file_name = f'{dataset_name}.mpack.gz'
            mpack_file_path: str = os.path.join(experiment.path, f'{dataset_name}.mpack.gz')
            if os.path.exists(mpack_file_path):
                file_size = os.path.getsize(mpack_file_path)
                file_size_mb = file_size / (1024 * 1024)
                click.secho(f'   uploading "{mpack_file_name}" [{file_size_mb:.1f} MB]')
                self.file_share.upload(mpack_file_name, mpack_file_path)

            csv_file_name = f'{dataset_name}.csv.gz'
            csv_file_path: str = os.path.join(experiment.path, f'{dataset_name}.csv.gz')
            if os.path.exists(csv_file_path):
                file_size = os.path.getsize(csv_file_path)
                file_size_mb = file_size / (1024 * 1024)
                click.secho(f'   uploading "{csv_file_name}" [{file_size_mb:.1f} MB]')
                self.file_share.upload(csv_file_name, csv_file_path)

        click.secho('✅ datasets uploaded', fg='green')
        click.secho()

    ## == DOCS COMMAND GROUP ==
    # This command group is used to manage the documentation of the package.
    
    @click.group('docs', help='Commands for managing the documentation of the package.')
    @click.pass_obj
    def docs_group(self,
                     ) -> None:
          """
          This command group contains commands that can be used to manage the documentation of the package.
          This includes commands to build the documentation, open it in a browser, or edit it.
          """
          pass
      
    @click.command('collect-datasets', help='Compiles the datasets for the documentation from the metadata file.')
    @click.option('--metadata-path', help='The path to the metadata file to use for compiling the datasets.',
                  default=METADATA_PATH, show_default=True)
    @click.option('--output-path', help='The path to the output file where the compiled datasets will be written.',
                  default=os.path.join(DOCS_PATH, 'datasets.md'), show_default=True)
    @click.pass_obj
    def docs_collect_datasets_command(self,
                                      metadata_path: str,
                                      output_path: str,
                                      ) -> None:
        """
        This command will dynamically create the overview of the datasets that are available in the 
        remote file share server using the metadata.yml file. By default, this will collect the dataset information 
        from the results of the experiments that have been executed the most recently.
        """
        
        ## -- Getting Metadata --
        # This dictionary contains all the metadata information stored on the remote file share server.
        click.secho('fetching metadata from remote fileshare server...')
        metadata: dict = self.file_share.fetch_metadata(force=True)
        
        ## -- Creating Table --
        # Using prettytable to create the table in the markdown format
        click.secho('compiling datasets overview...')
        table = PrettyTable()
        table.align = 'l'
        table.field_names = [
            "Name",
            "Description",
            "No. Elements",
            "Target Type",
        ]
        for dataset_name, dataset_info in metadata['datasets'].items():
            table.add_row([
                f'**{dataset_name}**',
                dataset_info.get('verbose', '-'),
                str(dataset_info.get('compounds', 0)),
                ', '.join(dataset_info.get('target_type', [])),
            ])
            
        # This is important to set the style to markdown so that the string that is created from this is 
        # actually a valid markdown syntax.
        table.set_style(TableStyle.MARKDOWN)
        datasets_table: str = table.get_string()
        
        ## -- Writing the Docs File --
        # Here we use the jinja2 template to actually render the content of the documentation file 
        # mainly centered around the just created table of the datasets.
        click.secho('writing datasets overview to file...')
        template = TEMPLATE_ENV.get_template('docs_dataset.md.j2')
        content = template.render(
            file_share_url=self.file_share.url,
            datasets_table=datasets_table,
        )
        with open(output_path, 'w') as file:
            file.write(content)
        
        click.secho('✅ datasets overview written to file', fg='green')
        click.secho('')


@click.group(cls=CLI)
@click.option("-v", "--version", is_flag=True, help='Print the version of the package')
@click.option('--no-verify', is_flag=True, help='turn off ssl verification for the fileshare connection')
@click.pass_context
def cli(ctx: Any,
        version: bool,
        no_verify: bool,
        ) -> None:

    # 07.06.24
    # This is actually required to make the CLI class work as it is currently implemented.
    # The problem is that all the commands which are also methods need to be decorated with the 
    # @pass_obj in the position of the "self" argument and here we set the object to be passed 
    # as the CLI instance to make that happen. 
    ctx.obj = ctx.command
    ctx.obj.file_share.verfiy = not no_verify

    if version:
        version = get_version()
        click.echo(Text(version, style=Style(bold=True)))
        sys.exit(0)
        
    # 07.06.24
    # There was previously a bug here which caused the help text to be rendered for every 
    # command. After adding the additional condition "not ctx.invoked_subcommand" this was 
    # fixed.
    
    # This section will implement the small convenience feature that if no arguments are passed 
    # to the CLI at all, it will simply show the help information.
    elif not ctx.args and not ctx.invoked_subcommand:
        ctx.command.format_help(ctx, ctx.formatter_class())



if __name__ == "__main__":
    cli()  # pragma: no cover


import io
from typing import List, Dict
import os
import gzip
import shutil
import requests
import pandas as pd

from rdkit import Chem
from pycomex.functional.experiment import Experiment
from pycomex.utils import folder_path, file_namespace


# :param DATASET_NAME:
#       This string determines the name of the message pack dataset file that is then 
#       stored into the "results" folder of the experiment as the result of the 
#       processing process. The corresponding file extensions will be added 
#       automatically.
DATASET_NAME: str = 'compas_2x'
# :param METADATA:
#       A dictionary which will be used as the basis for the metadata that will be added 
#       as additional information to the file share server.
METADATA: dict = {
    'description': (
        'This dataset is part of the COMputational database of Polycyclic Aromatic Systems (COMPAS) Project. '
        'It contains geometries and properties of cata-condensed poly(hetero)cyclic aromatic molecules, '
        'calculated at the GFN1-xTB level. The dataset includes approximately 500k molecules comprising 11 types '
        'of aromatic and antiaromatic building blocks, representing a highly diverse chemical space. '
        'Various electronic properties such as HOMO-LUMO gap, adiabatic ionization potential, and adiabatic electron '
        'affinity are provided. Additionally, the dataset is benchmarked against a ~50k dataset calculated at the '
        'CAM-B3LYP-D3BJ/def2-SVP level, with a fitting scheme developed to correct the xTB values to higher accuracy.'
    ),
    'target_type': ['Regression'],
    'tags': ['SMILES', 'Molecules', 'Quantum Chemistry', 'Molecular Properties'],
    'sources': [
        'https://chemrxiv.org/engage/chemrxiv/article-details/64bf8dd7b053dad33ad856cf',
        'https://gitlab.com/porannegroup/compas/-/tree/main/COMPAS-1?ref_type=heads',
    ],
    'target_descriptions': {
        0: 'homo - energy of the highest molecular orbit (HOMO) in electron volt (eV)',
        1: 'lumo - energy of the lowest unoccupied molecular orbit (LUMO) in electron volt (eV)',
        2: 'gap - energy gap between HOMO and LUMO in electron volt (eV)',
        3: 'energy - total energy of the molecule in electron volt (eV)',
        4: 'nfod - fractional occupation density',
        5: 'rings - number of rings in the molecule',
        6: 'dispersion - dispersion energy in Hartree (Eh)',
        7: 'aip - adiabatic ionization potential in Hartree (Eh)',
        8: 'aea - adiabatic electron affinity in Hartree (Eh)',
    }
}

__TESTING__ = False

experiment = Experiment.extend(
    'create_graph_datasets.py',
    base_path=folder_path(__file__),
    namespace=file_namespace(__file__),
    glob=globals(),
)


@experiment.hook('add_graph_metadata', default=False, replace=True)
def add_graph_metadata(e: Experiment, data: dict, graph: dict) -> dict:
    """
    We add the compound id for identification and the molecular weight
    """
    pass #graph['graph_id'] = data['ID']


@experiment.hook('load_dataset', default=False, replace=True)
def load_dataset(e: Experiment) -> Dict[int, dict]:
    
    # For the COMPAS 2 dataset the raw file is provided in the form of a CSV file that is hosted on Gitlab
    # here we download that file and then open the content as a pandas dataframe.
    e.log('downloading raw csv file...')
    raw_url = 'https://gitlab.com/porannegroup/compas/-/raw/main/COMPAS-2/compas-2x.csv'
    response = requests.get(raw_url, verify=False)
    response.raise_for_status()  # Ensure the request was successful
    csv_file = io.StringIO(response.text)
    df = pd.read_csv(csv_file, quotechar='"')
    
    print(df.head())
    
    compressed_path = os.path.join(e.path, f'{e.DATASET_NAME}.csv.gz')
    with gzip.open(compressed_path, mode='wb') as compressed_file:
        shutil.copyfileobj(csv_file, compressed_file)
    
    #df = load_smiles_dataset('qm9_smiles')

    columns = [
        'homo',
        'lumo',
        'gap',
        'energy',
        'nfod',
        'rings',
        'dispersion',
        'aip',
        'aea',
    ]

    dataset: Dict[int, dict] = {}
    index: int = 0
    for c, data in enumerate(df.to_dict('records')):
        data['smiles'] = data['smiles']
        
        # == MOLECULE FILTERS ==
        # We don't want to use compounds with '.' in the smiles (separate molecules)
        if '.' in data['smiles']:
            continue
        
        # We don't want to use compounds that only consist of a single atom
        mol = Chem.MolFromSmiles(data['smiles'])
        if not mol:
            continue
        
        # We also don't want to accept "molecules" that are essentially just individual atoms
        if len(mol.GetAtoms()) < 2:
            continue
        
        # == TARGETS ==
        # In this dataset we only have one target
        data['targets'] = [data[key] for key in columns]
        dataset[index] = data
        
        index += 1
        
        if c % 1000 == 0:
            print(f' * processed {c} records')

    return dataset

experiment.run_if_main()



import os
import tempfile
import yaml
from typing import List, Tuple

from chem_mat_data.cache import Cache
from .utils import get_mock_dataset_path


class TestCache:
    
    def test_add_dataset_basically_works(self):
        """
        the "add_dataset" method should copy the dataset to the specified cache folder and 
        also create a metadata file.
        """   
        with tempfile.TemporaryDirectory() as path:
            
            assert len(os.listdir(path)) == 0
            
            cache = Cache(path)
            metadata = {
                'num_elements': 100,
            }
            # This function will return the path to one dataset that is used for testing
            # (the mpack file version of that dataset)
            dataset_path = get_mock_dataset_path()
            cache.add_dataset('test', 'mpack', dataset_path, metadata)
            # This should create 2 files in the cache directory: the dataset itself and a metadata file
            assert len(os.listdir(path)) == 2
            
            files = os.listdir(path)
            print('cache files', files)
            assert 'test.mpack.yml' in files
            assert 'test.mpack.zip' in files
            
            # Then we can load that metadata file and inspect the content
            metadata_path = os.path.join(path, 'test.mpack.yml')
            with open(metadata_path) as file:
                metadata = yaml.load(file, Loader=yaml.FullLoader)
                
            print('metadata', metadata)
            assert isinstance(metadata, dict)
            assert 'num_elements' in metadata
            assert '_cache_filename' in metadata
            
    def test_list_datasets_basically_works(self):
        """
        The "list_datasets" method should return a list of tuples where each tuple contains the 
        name of the dataset and the type of the dataset that are currently stored in the cache folder.
        """
        with tempfile.TemporaryDirectory() as path:
            
            cache = Cache(path)
            cache.add_dataset('test_1', 'mpack', get_mock_dataset_path(), {'num_elements': 100})
            cache.add_dataset('test_2', 'mpack', get_mock_dataset_path(), {'num_elements': 200})
            
            datasets: List[Tuple[str, str]] = cache.list_datasets()
            assert len(datasets) == 2
            assert set(datasets) == set([('test_1', 'mpack'), ('test_2', 'mpack')])
            
    def test_contains_dataset_basically_works(self):
        """
        "contains_dataset" should return False at the start and then disable the dataset 
        after it has been added.
        """
        
        with tempfile.TemporaryDirectory() as path:
            
            cache = Cache(path)
            # Since the cache is empty, the contains function should return False
            assert not cache.contains_dataset('test', 'mpack')
            
            # Only after adding the dataset to the cache, the contains function should return True
            cache.add_dataset('test', 'mpack', get_mock_dataset_path(), {'num_elements': 100})
            assert cache.contains_dataset('test', 'mpack')
            
    def test_retrieve_dataset_basically_works(self):
        """
        Retrieve dataset should be able to retrieve a dataset from the cache and put it into the 
        destination folder using the original filename of the dataset when it was stored in the 
        cache.
        """
        # We setup two different temp folders here - one for the cache itself and the other to 
        # then retrieve the cache data into later on.
        with tempfile.TemporaryDirectory() as folder_path:
            with tempfile.TemporaryDirectory() as cache_path:
                
                cache = Cache(cache_path)
                dataset_path = get_mock_dataset_path()
                dataset_file_name = os.path.basename(dataset_path)
                cache.add_dataset('test', 'mpack', dataset_path, {'num_elements': 100})
                
                assert cache.contains_dataset('test', 'mpack')
                # Before retrieval the main folder should be empty
                assert len(os.listdir(folder_path)) == 0
                
                # Now we can retrieve the dataset from the cache
                cache.retrieve_dataset('test', 'mpack', folder_path)
                files = os.listdir(folder_path)
                print(files)
                assert len(files) == 1
                # Now the name of the reconstructed dataset should not actually be the name that 
                # we assigned to it in the cache but the original file name that was put into the 
                # dataset!
                assert dataset_file_name in files
                


import io
from typing import List, Dict
import os
import requests
import pandas as pd
import gzip
import shutil

from rdkit import Chem
from pycomex.functional.experiment import Experiment
from pycomex.utils import folder_path, file_namespace


# :param DATASET_NAME:
#       This string determines the name of the message pack dataset file that is then 
#       stored into the "results" folder of the experiment as the result of the 
#       processing process. The corresponding file extensions will be added 
#       automatically.
DATASET_NAME: str = 'compas_1x'
# :param METADATA:
#       A dictionary which will be used as the basis for the metadata that will be added 
#       as additional information to the file share server.
METADATA: dict = {
    'description': (
        'The COMPAS Project provides a computational database of polycyclic aromatic systems, '
        'including optimized ground-state structures and molecular properties of approximately '
        '34k and 9k cata-condensed polybenzenoid hydrocarbons. This dataset supports the investigation '
        'of structure-property relationships and the design of novel functional compounds, with data '
        'generated at the GFN2-xTB and B3LYP-D3BJ/def2-SVP levels.'
    ),
    'target_type': ['Regression'],
    'tags': ['SMILES', 'Molecules', 'Quantum Chemistry', 'Molecular Properties'],
    'sources': [
        'https://pubs.acs.org/doi/full/10.1021/acs.jcim.2c00503',
        'https://gitlab.com/porannegroup/compas/-/tree/main/COMPAS-1?ref_type=heads',
    ],
    'target_descriptions': {
        0: 'HOMO_eV - energy of the highest molecular orbit (HOMO) in electron volt (eV)',
        1: 'LUMO_eV - energy of the lowest unoccupied molecular orbit (LUMO) in electron volt (eV)',
        2: 'GAP_eV - energy gap between HOMO and LUMO in electron volt (eV)',
        3: 'Dipmom_Debye - dipole moment in Debye',
        4: 'Etot_eV - total energy of the molecule in electron volt (eV)',
        5: 'aEA_eV - adiabatic electron affinity in electron volt (eV)',
        6: 'aIP_eV - adiabatic ionization potential in electron volt (eV)',
        7: 'NFOD - fractional occupation density',
        8: 'n_rings - number of rings in the molecule',
    }
}

__TESTING__ = False

experiment = Experiment.extend(
    'create_graph_datasets.py',
    base_path=folder_path(__file__),
    namespace=file_namespace(__file__),
    glob=globals(),
)


@experiment.hook('add_graph_metadata', default=False, replace=True)
def add_graph_metadata(e: Experiment, data: dict, graph: dict) -> dict:
    """
    We add the compound id for identification and the molecular weight
    """
    pass #graph['graph_id'] = data['ID']


@experiment.hook('load_dataset', default=False, replace=True)
def load_dataset(e: Experiment) -> Dict[int, dict]:
    
    e.log('downloading raw csv file...')
    raw_url = 'https://gitlab.com/porannegroup/compas/-/raw/main/COMPAS-1/compas-1x.csv?ref_type=heads&inline=false'
    response = requests.get(raw_url, verify=False)
    csv_file = io.StringIO(response.text)
    df = pd.read_csv(csv_file)
    
    print(df.head())
    
    compressed_path = os.path.join(e.path, f'{e.DATASET_NAME}.csv.gz')
    with gzip.open(compressed_path, mode='wb') as compressed_file:
        shutil.copyfileobj(csv_file, compressed_file)

    columns = [
        'HOMO_eV',
        'LUMO_eV',
        'GAP_eV',
        'Dipmom_Debye',
        'Etot_eV',
        'aEA_eV',
        'aIP_eV',
        'NFOD',
        'n_rings',
    ]

    dataset: Dict[int, dict] = {}
    index: int = 0
    for data in df.to_dict('records'):
        data['smiles'] = data['smiles']
        
        # == MOLECULE FILTERS ==
        # We don't want to use compounds with '.' in the smiles (separate molecules)
        if '.' in data['smiles']:
            continue
        
        # We don't want to use compounds that only consist of a single atom
        mol = Chem.MolFromSmiles(data['smiles'])
        if not mol:
            continue
        
        # We also don't want to accept "molecules" that are essentially just individual atoms
        if len(mol.GetAtoms()) < 2:
            continue
        
        # == TARGETS ==
        # In this dataset we only have one target
        data['targets'] = [data[key] for key in columns]
        dataset[index] = data
        
        index += 1

    return dataset

experiment.run_if_main()



import rdkit.Chem as Chem
from typing import List, Dict
import pandas as pd

from pycomex.functional.experiment import Experiment
from pycomex.utils import folder_path, file_namespace

from chem_mat_data import load_smiles_dataset

DATASET_NAME: str = 'toxcast'

__TESTING__ = False

experiment = Experiment.extend(
    'create_graph_datasets.py',
    base_path=folder_path(__file__),
    namespace=file_namespace(__file__),
    glob=globals(),
)

@experiment.hook('add_graph_metadata', default=False, replace=True)
def add_graph_metadata(e: Experiment, data: dict, graph: dict) -> dict:
    """ 
    No metadata
    """
    pass

@experiment.hook('load_dataset', default=False, replace=True)
def load_dataset(e: Experiment) -> Dict[int, dict]:
    df = load_smiles_dataset('toxcast')

    # We have 617 targets...
    columns = ["ACEA_T47D_80hr_Negative",
        "ACEA_T47D_80hr_Positive",
        "APR_HepG2_CellCycleArrest_24h_dn",
        "APR_HepG2_CellCycleArrest_24h_up",
        "APR_HepG2_CellCycleArrest_72h_dn",
        "APR_HepG2_CellLoss_24h_dn",
        "APR_HepG2_CellLoss_72h_dn",
        "APR_HepG2_MicrotubuleCSK_24h_dn",
        "APR_HepG2_MicrotubuleCSK_24h_up",
        "APR_HepG2_MicrotubuleCSK_72h_dn",
        "APR_HepG2_MicrotubuleCSK_72h_up",
        "APR_HepG2_MitoMass_24h_dn",
        "APR_HepG2_MitoMass_24h_up",
        "APR_HepG2_MitoMass_72h_dn",
        "APR_HepG2_MitoMass_72h_up",
        "APR_HepG2_MitoMembPot_1h_dn",
        "APR_HepG2_MitoMembPot_24h_dn",
        "APR_HepG2_MitoMembPot_72h_dn",
        "APR_HepG2_MitoticArrest_24h_up",
        "APR_HepG2_MitoticArrest_72h_up",
        "APR_HepG2_NuclearSize_24h_dn",
        "APR_HepG2_NuclearSize_72h_dn",
        "APR_HepG2_NuclearSize_72h_up",
        "APR_HepG2_OxidativeStress_24h_up",
        "APR_HepG2_OxidativeStress_72h_up",
        "APR_HepG2_StressKinase_1h_up",
        "APR_HepG2_StressKinase_24h_up",
        "APR_HepG2_StressKinase_72h_up",
        "APR_HepG2_p53Act_24h_up",
        "APR_HepG2_p53Act_72h_up",
        "APR_Hepat_Apoptosis_24hr_up",
        "APR_Hepat_Apoptosis_48hr_up",
        "APR_Hepat_CellLoss_24hr_dn",
        "APR_Hepat_CellLoss_48hr_dn",
        "APR_Hepat_DNADamage_24hr_up",
        "APR_Hepat_DNADamage_48hr_up",
        "APR_Hepat_DNATexture_24hr_up",
        "APR_Hepat_DNATexture_48hr_up",
        "APR_Hepat_MitoFxnI_1hr_dn",
        "APR_Hepat_MitoFxnI_24hr_dn",
        "APR_Hepat_MitoFxnI_48hr_dn",
        "APR_Hepat_NuclearSize_24hr_dn",
        "APR_Hepat_NuclearSize_48hr_dn",
        "APR_Hepat_Steatosis_24hr_up",
        "APR_Hepat_Steatosis_48hr_up",
        "ATG_AP_1_CIS_dn",
        "ATG_AP_1_CIS_up",
        "ATG_AP_2_CIS_dn",
        "ATG_AP_2_CIS_up",
        "ATG_AR_TRANS_dn",
        "ATG_AR_TRANS_up",
        "ATG_Ahr_CIS_dn",
        "ATG_Ahr_CIS_up",
        "ATG_BRE_CIS_dn",
        "ATG_BRE_CIS_up",
        "ATG_CAR_TRANS_dn",
        "ATG_CAR_TRANS_up",
        "ATG_CMV_CIS_dn",
        "ATG_CMV_CIS_up",
        "ATG_CRE_CIS_dn",
        "ATG_CRE_CIS_up",
        "ATG_C_EBP_CIS_dn",
        "ATG_C_EBP_CIS_up",
        "ATG_DR4_LXR_CIS_dn",
        "ATG_DR4_LXR_CIS_up",
        "ATG_DR5_CIS_dn",
        "ATG_DR5_CIS_up",
        "ATG_E2F_CIS_dn",
        "ATG_E2F_CIS_up",
        "ATG_EGR_CIS_up",
        "ATG_ERE_CIS_dn",
        "ATG_ERE_CIS_up",
        "ATG_ERRa_TRANS_dn",
        "ATG_ERRg_TRANS_dn",
        "ATG_ERRg_TRANS_up",
        "ATG_ERa_TRANS_up",
        "ATG_E_Box_CIS_dn",
        "ATG_E_Box_CIS_up",
        "ATG_Ets_CIS_dn",
        "ATG_Ets_CIS_up",
        "ATG_FXR_TRANS_up",
        "ATG_FoxA2_CIS_dn",
        "ATG_FoxA2_CIS_up",
        "ATG_FoxO_CIS_dn",
        "ATG_FoxO_CIS_up",
        "ATG_GAL4_TRANS_dn",
        "ATG_GATA_CIS_dn",
        "ATG_GATA_CIS_up",
        "ATG_GLI_CIS_dn",
        "ATG_GLI_CIS_up",
        "ATG_GRE_CIS_dn",
        "ATG_GRE_CIS_up",
        "ATG_GR_TRANS_dn",
        "ATG_GR_TRANS_up",
        "ATG_HIF1a_CIS_dn",
        "ATG_HIF1a_CIS_up",
        "ATG_HNF4a_TRANS_dn",
        "ATG_HNF4a_TRANS_up",
        "ATG_HNF6_CIS_dn",
        "ATG_HNF6_CIS_up",
        "ATG_HSE_CIS_dn",
        "ATG_HSE_CIS_up",
        "ATG_IR1_CIS_dn",
        "ATG_IR1_CIS_up",
        "ATG_ISRE_CIS_dn",
        "ATG_ISRE_CIS_up",
        "ATG_LXRa_TRANS_dn",
        "ATG_LXRa_TRANS_up",
        "ATG_LXRb_TRANS_dn",
        "ATG_LXRb_TRANS_up",
        "ATG_MRE_CIS_up",
        "ATG_M_06_TRANS_up",
        "ATG_M_19_CIS_dn",
        "ATG_M_19_TRANS_dn",
        "ATG_M_19_TRANS_up",
        "ATG_M_32_CIS_dn",
        "ATG_M_32_CIS_up",
        "ATG_M_32_TRANS_dn",
        "ATG_M_32_TRANS_up",
        "ATG_M_61_TRANS_up",
        "ATG_Myb_CIS_dn",
        "ATG_Myb_CIS_up",
        "ATG_Myc_CIS_dn",
        "ATG_Myc_CIS_up",
        "ATG_NFI_CIS_dn",
        "ATG_NFI_CIS_up",
        "ATG_NF_kB_CIS_dn",
        "ATG_NF_kB_CIS_up",
        "ATG_NRF1_CIS_dn",
        "ATG_NRF1_CIS_up",
        "ATG_NRF2_ARE_CIS_dn",
        "ATG_NRF2_ARE_CIS_up",
        "ATG_NURR1_TRANS_dn",
        "ATG_NURR1_TRANS_up",
        "ATG_Oct_MLP_CIS_dn",
        "ATG_Oct_MLP_CIS_up",
        "ATG_PBREM_CIS_dn",
        "ATG_PBREM_CIS_up",
        "ATG_PPARa_TRANS_dn",
        "ATG_PPARa_TRANS_up",
        "ATG_PPARd_TRANS_up",
        "ATG_PPARg_TRANS_up",
        "ATG_PPRE_CIS_dn",
        "ATG_PPRE_CIS_up",
        "ATG_PXRE_CIS_dn",
        "ATG_PXRE_CIS_up",
        "ATG_PXR_TRANS_dn",
        "ATG_PXR_TRANS_up",
        "ATG_Pax6_CIS_up",
        "ATG_RARa_TRANS_dn",
        "ATG_RARa_TRANS_up",
        "ATG_RARb_TRANS_dn",
        "ATG_RARb_TRANS_up",
        "ATG_RARg_TRANS_dn",
        "ATG_RARg_TRANS_up",
        "ATG_RORE_CIS_dn",
        "ATG_RORE_CIS_up",
        "ATG_RORb_TRANS_dn",
        "ATG_RORg_TRANS_dn",
        "ATG_RORg_TRANS_up",
        "ATG_RXRa_TRANS_dn",
        "ATG_RXRa_TRANS_up",
        "ATG_RXRb_TRANS_dn",
        "ATG_RXRb_TRANS_up",
        "ATG_SREBP_CIS_dn",
        "ATG_SREBP_CIS_up",
        "ATG_STAT3_CIS_dn",
        "ATG_STAT3_CIS_up",
        "ATG_Sox_CIS_dn",
        "ATG_Sox_CIS_up",
        "ATG_Sp1_CIS_dn",
        "ATG_Sp1_CIS_up",
        "ATG_TAL_CIS_dn",
        "ATG_TAL_CIS_up",
        "ATG_TA_CIS_dn",
        "ATG_TA_CIS_up",
        "ATG_TCF_b_cat_CIS_dn",
        "ATG_TCF_b_cat_CIS_up",
        "ATG_TGFb_CIS_dn",
        "ATG_TGFb_CIS_up",
        "ATG_THRa1_TRANS_dn",
        "ATG_THRa1_TRANS_up",
        "ATG_VDRE_CIS_dn",
        "ATG_VDRE_CIS_up",
        "ATG_VDR_TRANS_dn",
        "ATG_VDR_TRANS_up",
        "ATG_XTT_Cytotoxicity_up",
        "ATG_Xbp1_CIS_dn",
        "ATG_Xbp1_CIS_up",
        "ATG_p53_CIS_dn",
        "ATG_p53_CIS_up",
        "BSK_3C_Eselectin_down",
        "BSK_3C_HLADR_down",
        "BSK_3C_ICAM1_down",
        "BSK_3C_IL8_down",
        "BSK_3C_MCP1_down",
        "BSK_3C_MIG_down",
        "BSK_3C_Proliferation_down",
        "BSK_3C_SRB_down",
        "BSK_3C_Thrombomodulin_down",
        "BSK_3C_Thrombomodulin_up",
        "BSK_3C_TissueFactor_down",
        "BSK_3C_TissueFactor_up",
        "BSK_3C_VCAM1_down",
        "BSK_3C_Vis_down",
        "BSK_3C_uPAR_down",
        "BSK_4H_Eotaxin3_down",
        "BSK_4H_MCP1_down",
        "BSK_4H_Pselectin_down",
        "BSK_4H_Pselectin_up",
        "BSK_4H_SRB_down",
        "BSK_4H_VCAM1_down",
        "BSK_4H_VEGFRII_down",
        "BSK_4H_uPAR_down",
        "BSK_4H_uPAR_up",
        "BSK_BE3C_HLADR_down",
        "BSK_BE3C_IL1a_down",
        "BSK_BE3C_IP10_down",
        "BSK_BE3C_MIG_down",
        "BSK_BE3C_MMP1_down",
        "BSK_BE3C_MMP1_up",
        "BSK_BE3C_PAI1_down",
        "BSK_BE3C_SRB_down",
        "BSK_BE3C_TGFb1_down",
        "BSK_BE3C_tPA_down",
        "BSK_BE3C_uPAR_down",
        "BSK_BE3C_uPAR_up",
        "BSK_BE3C_uPA_down",
        "BSK_CASM3C_HLADR_down",
        "BSK_CASM3C_IL6_down",
        "BSK_CASM3C_IL6_up",
        "BSK_CASM3C_IL8_down",
        "BSK_CASM3C_LDLR_down",
        "BSK_CASM3C_LDLR_up",
        "BSK_CASM3C_MCP1_down",
        "BSK_CASM3C_MCP1_up",
        "BSK_CASM3C_MCSF_down",
        "BSK_CASM3C_MCSF_up",
        "BSK_CASM3C_MIG_down",
        "BSK_CASM3C_Proliferation_down",
        "BSK_CASM3C_Proliferation_up",
        "BSK_CASM3C_SAA_down",
        "BSK_CASM3C_SAA_up",
        "BSK_CASM3C_SRB_down",
        "BSK_CASM3C_Thrombomodulin_down",
        "BSK_CASM3C_Thrombomodulin_up",
        "BSK_CASM3C_TissueFactor_down",
        "BSK_CASM3C_VCAM1_down",
        "BSK_CASM3C_VCAM1_up",
        "BSK_CASM3C_uPAR_down",
        "BSK_CASM3C_uPAR_up",
        "BSK_KF3CT_ICAM1_down",
        "BSK_KF3CT_IL1a_down",
        "BSK_KF3CT_IP10_down",
        "BSK_KF3CT_IP10_up",
        "BSK_KF3CT_MCP1_down",
        "BSK_KF3CT_MCP1_up",
        "BSK_KF3CT_MMP9_down",
        "BSK_KF3CT_SRB_down",
        "BSK_KF3CT_TGFb1_down",
        "BSK_KF3CT_TIMP2_down",
        "BSK_KF3CT_uPA_down",
        "BSK_LPS_CD40_down",
        "BSK_LPS_Eselectin_down",
        "BSK_LPS_Eselectin_up",
        "BSK_LPS_IL1a_down",
        "BSK_LPS_IL1a_up",
        "BSK_LPS_IL8_down",
        "BSK_LPS_IL8_up",
        "BSK_LPS_MCP1_down",
        "BSK_LPS_MCSF_down",
        "BSK_LPS_PGE2_down",
        "BSK_LPS_PGE2_up",
        "BSK_LPS_SRB_down",
        "BSK_LPS_TNFa_down",
        "BSK_LPS_TNFa_up",
        "BSK_LPS_TissueFactor_down",
        "BSK_LPS_TissueFactor_up",
        "BSK_LPS_VCAM1_down",
        "BSK_SAg_CD38_down",
        "BSK_SAg_CD40_down",
        "BSK_SAg_CD69_down",
        "BSK_SAg_Eselectin_down",
        "BSK_SAg_Eselectin_up",
        "BSK_SAg_IL8_down",
        "BSK_SAg_IL8_up",
        "BSK_SAg_MCP1_down",
        "BSK_SAg_MIG_down",
        "BSK_SAg_PBMCCytotoxicity_down",
        "BSK_SAg_PBMCCytotoxicity_up",
        "BSK_SAg_Proliferation_down",
        "BSK_SAg_SRB_down",
        "BSK_hDFCGF_CollagenIII_down",
        "BSK_hDFCGF_EGFR_down",
        "BSK_hDFCGF_EGFR_up",
        "BSK_hDFCGF_IL8_down",
        "BSK_hDFCGF_IP10_down",
        "BSK_hDFCGF_MCSF_down",
        "BSK_hDFCGF_MIG_down",
        "BSK_hDFCGF_MMP1_down",
        "BSK_hDFCGF_MMP1_up",
        "BSK_hDFCGF_PAI1_down",
        "BSK_hDFCGF_Proliferation_down",
        "BSK_hDFCGF_SRB_down",
        "BSK_hDFCGF_TIMP1_down",
        "BSK_hDFCGF_VCAM1_down",
        "CEETOX_H295R_11DCORT_dn",
        "CEETOX_H295R_ANDR_dn",
        "CEETOX_H295R_CORTISOL_dn",
        "CEETOX_H295R_DOC_dn",
        "CEETOX_H295R_DOC_up",
        "CEETOX_H295R_ESTRADIOL_dn",
        "CEETOX_H295R_ESTRADIOL_up",
        "CEETOX_H295R_ESTRONE_dn",
        "CEETOX_H295R_ESTRONE_up",
        "CEETOX_H295R_OHPREG_up",
        "CEETOX_H295R_OHPROG_dn",
        "CEETOX_H295R_OHPROG_up",
        "CEETOX_H295R_PROG_up",
        "CEETOX_H295R_TESTO_dn",
        "CLD_ABCB1_48hr",
        "CLD_ABCG2_48hr",
        "CLD_CYP1A1_24hr",
        "CLD_CYP1A1_48hr",
        "CLD_CYP1A1_6hr",
        "CLD_CYP1A2_24hr",
        "CLD_CYP1A2_48hr",
        "CLD_CYP1A2_6hr",
        "CLD_CYP2B6_24hr",
        "CLD_CYP2B6_48hr",
        "CLD_CYP2B6_6hr",
        "CLD_CYP3A4_24hr",
        "CLD_CYP3A4_48hr",
        "CLD_CYP3A4_6hr",
        "CLD_GSTA2_48hr",
        "CLD_SULT2A_24hr",
        "CLD_SULT2A_48hr",
        "CLD_UGT1A1_24hr",
        "CLD_UGT1A1_48hr",
        "NCCT_HEK293T_CellTiterGLO",
        "NCCT_QuantiLum_inhib_2_dn",
        "NCCT_QuantiLum_inhib_dn",
        "NCCT_TPO_AUR_dn",
        "NCCT_TPO_GUA_dn",
        "NHEERL_ZF_144hpf_TERATOSCORE_up",
        "NVS_ADME_hCYP19A1",
        "NVS_ADME_hCYP1A1",
        "NVS_ADME_hCYP1A2",
        "NVS_ADME_hCYP2A6",
        "NVS_ADME_hCYP2B6",
        "NVS_ADME_hCYP2C19",
        "NVS_ADME_hCYP2C9",
        "NVS_ADME_hCYP2D6",
        "NVS_ADME_hCYP3A4",
        "NVS_ADME_hCYP4F12",
        "NVS_ADME_rCYP2C12",
        "NVS_ENZ_hAChE",
        "NVS_ENZ_hAMPKa1",
        "NVS_ENZ_hAurA",
        "NVS_ENZ_hBACE",
        "NVS_ENZ_hCASP5",
        "NVS_ENZ_hCK1D",
        "NVS_ENZ_hDUSP3",
        "NVS_ENZ_hES",
        "NVS_ENZ_hElastase",
        "NVS_ENZ_hFGFR1",
        "NVS_ENZ_hGSK3b",
        "NVS_ENZ_hMMP1",
        "NVS_ENZ_hMMP13",
        "NVS_ENZ_hMMP2",
        "NVS_ENZ_hMMP3",
        "NVS_ENZ_hMMP7",
        "NVS_ENZ_hMMP9",
        "NVS_ENZ_hPDE10",
        "NVS_ENZ_hPDE4A1",
        "NVS_ENZ_hPDE5",
        "NVS_ENZ_hPI3Ka",
        "NVS_ENZ_hPTEN",
        "NVS_ENZ_hPTPN11",
        "NVS_ENZ_hPTPN12",
        "NVS_ENZ_hPTPN13",
        "NVS_ENZ_hPTPN9",
        "NVS_ENZ_hPTPRC",
        "NVS_ENZ_hSIRT1",
        "NVS_ENZ_hSIRT2",
        "NVS_ENZ_hTrkA",
        "NVS_ENZ_hVEGFR2",
        "NVS_ENZ_oCOX1",
        "NVS_ENZ_oCOX2",
        "NVS_ENZ_rAChE",
        "NVS_ENZ_rCNOS",
        "NVS_ENZ_rMAOAC",
        "NVS_ENZ_rMAOAP",
        "NVS_ENZ_rMAOBC",
        "NVS_ENZ_rMAOBP",
        "NVS_ENZ_rabI2C",
        "NVS_GPCR_bAdoR_NonSelective",
        "NVS_GPCR_bDR_NonSelective",
        "NVS_GPCR_g5HT4",
        "NVS_GPCR_gH2",
        "NVS_GPCR_gLTB4",
        "NVS_GPCR_gLTD4",
        "NVS_GPCR_gMPeripheral_NonSelective",
        "NVS_GPCR_gOpiateK",
        "NVS_GPCR_h5HT2A",
        "NVS_GPCR_h5HT5A",
        "NVS_GPCR_h5HT6",
        "NVS_GPCR_h5HT7",
        "NVS_GPCR_hAT1",
        "NVS_GPCR_hAdoRA1",
        "NVS_GPCR_hAdoRA2a",
        "NVS_GPCR_hAdra2A",
        "NVS_GPCR_hAdra2C",
        "NVS_GPCR_hAdrb1",
        "NVS_GPCR_hAdrb2",
        "NVS_GPCR_hAdrb3",
        "NVS_GPCR_hDRD1",
        "NVS_GPCR_hDRD2s",
        "NVS_GPCR_hDRD4.4",
        "NVS_GPCR_hH1",
        "NVS_GPCR_hLTB4_BLT1",
        "NVS_GPCR_hM1",
        "NVS_GPCR_hM2",
        "NVS_GPCR_hM3",
        "NVS_GPCR_hM4",
        "NVS_GPCR_hNK2",
        "NVS_GPCR_hOpiate_D1",
        "NVS_GPCR_hOpiate_mu",
        "NVS_GPCR_hTXA2",
        "NVS_GPCR_p5HT2C",
        "NVS_GPCR_r5HT1_NonSelective",
        "NVS_GPCR_r5HT_NonSelective",
        "NVS_GPCR_rAdra1B",
        "NVS_GPCR_rAdra1_NonSelective",
        "NVS_GPCR_rAdra2_NonSelective",
        "NVS_GPCR_rAdrb_NonSelective",
        "NVS_GPCR_rNK1",
        "NVS_GPCR_rNK3",
        "NVS_GPCR_rOpiate_NonSelective",
        "NVS_GPCR_rOpiate_NonSelectiveNa",
        "NVS_GPCR_rSST",
        "NVS_GPCR_rTRH",
        "NVS_GPCR_rV1",
        "NVS_GPCR_rabPAF",
        "NVS_GPCR_rmAdra2B",
        "NVS_IC_hKhERGCh",
        "NVS_IC_rCaBTZCHL",
        "NVS_IC_rCaDHPRCh_L",
        "NVS_IC_rNaCh_site2",
        "NVS_LGIC_bGABARa1",
        "NVS_LGIC_h5HT3",
        "NVS_LGIC_hNNR_NBungSens",
        "NVS_LGIC_rGABAR_NonSelective",
        "NVS_LGIC_rNNR_BungSens",
        "NVS_MP_hPBR",
        "NVS_MP_rPBR",
        "NVS_NR_bER",
        "NVS_NR_bPR",
        "NVS_NR_cAR",
        "NVS_NR_hAR",
        "NVS_NR_hCAR_Antagonist",
        "NVS_NR_hER",
        "NVS_NR_hFXR_Agonist",
        "NVS_NR_hFXR_Antagonist",
        "NVS_NR_hGR",
        "NVS_NR_hPPARa",
        "NVS_NR_hPPARg",
        "NVS_NR_hPR",
        "NVS_NR_hPXR",
        "NVS_NR_hRAR_Antagonist",
        "NVS_NR_hRARa_Agonist",
        "NVS_NR_hTRa_Antagonist",
        "NVS_NR_mERa",
        "NVS_NR_rAR",
        "NVS_NR_rMR",
        "NVS_OR_gSIGMA_NonSelective",
        "NVS_TR_gDAT",
        "NVS_TR_hAdoT",
        "NVS_TR_hDAT",
        "NVS_TR_hNET",
        "NVS_TR_hSERT",
        "NVS_TR_rNET",
        "NVS_TR_rSERT",
        "NVS_TR_rVMAT2",
        "OT_AR_ARELUC_AG_1440",
        "OT_AR_ARSRC1_0480",
        "OT_AR_ARSRC1_0960",
        "OT_ER_ERaERa_0480",
        "OT_ER_ERaERa_1440",
        "OT_ER_ERaERb_0480",
        "OT_ER_ERaERb_1440",
        "OT_ER_ERbERb_0480",
        "OT_ER_ERbERb_1440",
        "OT_ERa_EREGFP_0120",
        "OT_ERa_EREGFP_0480",
        "OT_FXR_FXRSRC1_0480",
        "OT_FXR_FXRSRC1_1440",
        "OT_NURR1_NURR1RXRa_0480",
        "OT_NURR1_NURR1RXRa_1440",
        "TOX21_ARE_BLA_Agonist_ch1",
        "TOX21_ARE_BLA_Agonist_ch2",
        "TOX21_ARE_BLA_agonist_ratio",
        "TOX21_ARE_BLA_agonist_viability",
        "TOX21_AR_BLA_Agonist_ch1",
        "TOX21_AR_BLA_Agonist_ch2",
        "TOX21_AR_BLA_Agonist_ratio",
        "TOX21_AR_BLA_Antagonist_ch1",
        "TOX21_AR_BLA_Antagonist_ch2",
        "TOX21_AR_BLA_Antagonist_ratio",
        "TOX21_AR_BLA_Antagonist_viability",
        "TOX21_AR_LUC_MDAKB2_Agonist",
        "TOX21_AR_LUC_MDAKB2_Antagonist",
        "TOX21_AR_LUC_MDAKB2_Antagonist2",
        "TOX21_AhR_LUC_Agonist",
        "TOX21_Aromatase_Inhibition",
        "TOX21_AutoFluor_HEK293_Cell_blue",
        "TOX21_AutoFluor_HEK293_Media_blue",
        "TOX21_AutoFluor_HEPG2_Cell_blue",
        "TOX21_AutoFluor_HEPG2_Cell_green",
        "TOX21_AutoFluor_HEPG2_Media_blue",
        "TOX21_AutoFluor_HEPG2_Media_green",
        "TOX21_ELG1_LUC_Agonist",
        "TOX21_ERa_BLA_Agonist_ch1",
        "TOX21_ERa_BLA_Agonist_ch2",
        "TOX21_ERa_BLA_Agonist_ratio",
        "TOX21_ERa_BLA_Antagonist_ch1",
        "TOX21_ERa_BLA_Antagonist_ch2",
        "TOX21_ERa_BLA_Antagonist_ratio",
        "TOX21_ERa_BLA_Antagonist_viability",
        "TOX21_ERa_LUC_BG1_Agonist",
        "TOX21_ERa_LUC_BG1_Antagonist",
        "TOX21_ESRE_BLA_ch1",
        "TOX21_ESRE_BLA_ch2",
        "TOX21_ESRE_BLA_ratio",
        "TOX21_ESRE_BLA_viability",
        "TOX21_FXR_BLA_Antagonist_ch1",
        "TOX21_FXR_BLA_Antagonist_ch2",
        "TOX21_FXR_BLA_agonist_ch2",
        "TOX21_FXR_BLA_agonist_ratio",
        "TOX21_FXR_BLA_antagonist_ratio",
        "TOX21_FXR_BLA_antagonist_viability",
        "TOX21_GR_BLA_Agonist_ch1",
        "TOX21_GR_BLA_Agonist_ch2",
        "TOX21_GR_BLA_Agonist_ratio",
        "TOX21_GR_BLA_Antagonist_ch2",
        "TOX21_GR_BLA_Antagonist_ratio",
        "TOX21_GR_BLA_Antagonist_viability",
        "TOX21_HSE_BLA_agonist_ch1",
        "TOX21_HSE_BLA_agonist_ch2",
        "TOX21_HSE_BLA_agonist_ratio",
        "TOX21_HSE_BLA_agonist_viability",
        "TOX21_MMP_ratio_down",
        "TOX21_MMP_ratio_up",
        "TOX21_MMP_viability",
        "TOX21_NFkB_BLA_agonist_ch1",
        "TOX21_NFkB_BLA_agonist_ch2",
        "TOX21_NFkB_BLA_agonist_ratio",
        "TOX21_NFkB_BLA_agonist_viability",
        "TOX21_PPARd_BLA_Agonist_viability",
        "TOX21_PPARd_BLA_Antagonist_ch1",
        "TOX21_PPARd_BLA_agonist_ch1",
        "TOX21_PPARd_BLA_agonist_ch2",
        "TOX21_PPARd_BLA_agonist_ratio",
        "TOX21_PPARd_BLA_antagonist_ratio",
        "TOX21_PPARd_BLA_antagonist_viability",
        "TOX21_PPARg_BLA_Agonist_ch1",
        "TOX21_PPARg_BLA_Agonist_ch2",
        "TOX21_PPARg_BLA_Agonist_ratio",
        "TOX21_PPARg_BLA_Antagonist_ch1",
        "TOX21_PPARg_BLA_antagonist_ratio",
        "TOX21_PPARg_BLA_antagonist_viability",
        "TOX21_TR_LUC_GH3_Agonist",
        "TOX21_TR_LUC_GH3_Antagonist",
        "TOX21_VDR_BLA_Agonist_viability",
        "TOX21_VDR_BLA_Antagonist_ch1",
        "TOX21_VDR_BLA_agonist_ch2",
        "TOX21_VDR_BLA_agonist_ratio",
        "TOX21_VDR_BLA_antagonist_ratio",
        "TOX21_VDR_BLA_antagonist_viability",
        "TOX21_p53_BLA_p1_ch1",
        "TOX21_p53_BLA_p1_ch2",
        "TOX21_p53_BLA_p1_ratio",
        "TOX21_p53_BLA_p1_viability",
        "TOX21_p53_BLA_p2_ch1",
        "TOX21_p53_BLA_p2_ch2",
        "TOX21_p53_BLA_p2_ratio",
        "TOX21_p53_BLA_p2_viability",
        "TOX21_p53_BLA_p3_ch1",
        "TOX21_p53_BLA_p3_ch2",
        "TOX21_p53_BLA_p3_ratio",
        "TOX21_p53_BLA_p3_viability",
        "TOX21_p53_BLA_p4_ch1",
        "TOX21_p53_BLA_p4_ch2",
        "TOX21_p53_BLA_p4_ratio",
        "TOX21_p53_BLA_p4_viability",
        "TOX21_p53_BLA_p5_ch1",
        "TOX21_p53_BLA_p5_ch2",
        "TOX21_p53_BLA_p5_ratio",
        "TOX21_p53_BLA_p5_viability",
        "Tanguay_ZF_120hpf_AXIS_up",
        "Tanguay_ZF_120hpf_ActivityScore",
        "Tanguay_ZF_120hpf_BRAI_up",
        "Tanguay_ZF_120hpf_CFIN_up",
        "Tanguay_ZF_120hpf_CIRC_up",
        "Tanguay_ZF_120hpf_EYE_up",
        "Tanguay_ZF_120hpf_JAW_up",
        "Tanguay_ZF_120hpf_MORT_up",
        "Tanguay_ZF_120hpf_OTIC_up",
        "Tanguay_ZF_120hpf_PE_up",
        "Tanguay_ZF_120hpf_PFIN_up",
        "Tanguay_ZF_120hpf_PIG_up",
        "Tanguay_ZF_120hpf_SNOU_up",
        "Tanguay_ZF_120hpf_SOMI_up",
        "Tanguay_ZF_120hpf_SWIM_up",
        "Tanguay_ZF_120hpf_TRUN_up",
        "Tanguay_ZF_120hpf_TR_up",
        "Tanguay_ZF_120hpf_YSE_up"]
    
    
    dataset: Dict[int, dict] = {}
    for index, data in enumerate(df.to_dict('records')):
        
        data['smiles'] = data['smiles']
        
        # === MOLECULE FILTERS ===
        # We don't want to use compounds with '.' in the smiles (separate molecules)
        if '.' in data['smiles']:
            continue
        
        # We don't want to use compounds that only consist of a single atom
        mol = Chem.MolFromSmiles(data['smiles'])
        if not mol:
            continue
        
        # We also don't want to accept "molecules" that are essentially just individual atoms
        if len(mol.GetAtoms()) < 2:
            continue
        
        # == TARGETS ==
        data['targets'] = [(0 if data[col] == 0 else 1) if pd.notna(data[col]) else -1 for col in columns]
        dataset[index] = data

    return dataset

experiment.run_if_main()




"""
This module contains the unittests that specifically test the bridges to the major graph 
neural network libraries such as ptorch geometric and jraph.

NOTE: The GNN libraries used here are not part of the packages main dependencies and have 
      to be installed separately.
"""
import time

import torch
import torch.nn as nn
from torch_geometric.data import Data
from torch_geometric.loader import DataLoader
from torch_geometric.nn.conv import GCNConv
from torch_geometric.nn.aggr import SumAggregation

import jax
import jraph

from chem_mat_data.testing import create_mock_graph
from chem_mat_data.testing import sample_mock_graphs
from chem_mat_data.main import pyg_from_graph
from chem_mat_data.main import pyg_data_list_from_graphs
from chem_mat_data.main import jraph_from_graph
from chem_mat_data.main import jraph_implicit_batch_from_graphs


# == TORCH/PYTORCH GEOMETRIC ==

class TestPytorchGeometric:
    
    def test_pyg_from_graph_basically_works(self):
        """
        ``pyg_from_graph`` should take a graph dict as input and turn it into a pyg Data object instance
        """
        num_nodes = 15
        num_node_attributes = 20
        num_edge_attributes = 10
        num_graph_labels = 2
        
        graph = create_mock_graph(
            num_nodes=num_nodes,
            num_node_attributes=num_node_attributes,
            num_edge_attributes=num_edge_attributes,
            num_graph_labels=num_graph_labels
        )
        data: Data = pyg_from_graph(graph)
        assert isinstance(data, Data)
        
        assert isinstance(data.x, torch.Tensor)
        assert data.x.shape == (num_nodes, num_node_attributes)
        
        assert isinstance(data.edge_attr, torch.Tensor)
        assert data.edge_attr.shape == (num_nodes, num_edge_attributes)
        
        assert isinstance(data.edge_index, torch.Tensor)
        assert data.edge_index.shape == (2, num_nodes)
        
        # additionally, since the mock graphs actually contain target labels, these target labels should 
        # also be converted to tensors and stored in the data object.
        assert isinstance(data.y, torch.Tensor)
        assert data.y.shape == (num_graph_labels, )
        
    def test_pyg_data_list_from_graphs_basically_works(self):
        """
        ``pyg_data_list_from_graphs`` should take a list of graph dicts as input and turn them into a list of
        pyg Data object instances which should then be usable to train a pyg gnn model.
        """
        num_graphs = 10
        num_node_range = (5, 10)
        num_node_attributes = 20
        num_edge_attributes = 10
        num_graph_labels = 3
        
        graphs = sample_mock_graphs(
            k=num_graphs,
            num_nodes_range=num_node_range,
            num_node_attributes=num_node_attributes,
            num_edge_attributes=num_edge_attributes,
            num_graph_labels=num_graph_labels
        )
        
        data_list = pyg_data_list_from_graphs(graphs)
        assert isinstance(data_list, list)
        assert len(data_list) == num_graphs
        
        for data, graph in zip(data_list, graphs):
            assert isinstance(data, Data)
            assert data.x.shape[0] == len(graph['node_indices'])

    def test_import_computational_overhead(self):
        """
        The function ``pyg_from_graph`` performs a local import for every call of the function 
        to convert an individual graph while the ``pyg_data_list_from_graphs`` function only does 
        this import once and then converts the graphs individually. This test assesses the 
        difference in computational time between the two implementations.
        """
        num_graphs = 1000
        num_nodes_range = (20, 50)
        num_node_attributes = 40
        num_edge_attributes = 20
        num_graph_labels = 1
        
        graphs = sample_mock_graphs(
            k=num_graphs,
            num_nodes_range=num_nodes_range,
            num_node_attributes=num_node_attributes,
            num_edge_attributes=num_edge_attributes,
            num_graph_labels=num_graph_labels,
        )
        
        # ~ individual graphs
        time_start = time.time()
        for graph in graphs:
            pyg_from_graph(graph)
            
        duration_single = time.time() - time_start

        # ~ batch processing
        time_start = time.time()
        pyg_data_list_from_graphs(graphs)
        
        duration_batch = time.time() - time_start
        
        print(f'num graphs: {num_graphs}')
        print(f'single: {duration_single*10e3/num_graphs:.4f}ms/graph - '
              f'batch: {duration_batch*10e3/num_graphs:.4f}ms/graph')
        print(f'single increase: {duration_single/duration_batch:.3f}x')
        
    def test_pyg_model_compatibility(self):
        """
        The converted pyg data object should be possible to be used as input to the forward pass 
        of a simple pyg model.
        """
        num_graphs = 10
        num_nodes_range = (5, 10)
        num_node_attributes = 10
        num_edge_attributes = 5
        num_graph_labels = 2
        
        graphs = sample_mock_graphs(
            k=num_graphs,
            num_nodes_range=num_nodes_range,
            num_node_attributes=num_node_attributes,
            num_edge_attributes=num_edge_attributes,
            num_graph_labels=num_graph_labels,
        )
        
        class Model(nn.Module):
            
            def __init__(self):
                super().__init__()
                self.gcn_1 = GCNConv(num_node_attributes, 16)
                self.gcn_2 = GCNConv(16, 8)
                self.gcn_3 = GCNConv(8, num_graph_labels)
                self.pool = SumAggregation()

            def forward(self, data: Data):
                x = data.x
                x = self.gcn_1(x, data.edge_index)
                x = self.gcn_2(x, data.edge_index)
                x = self.gcn_3(x, data.edge_index)
                x = self.pool(x, data.batch)
                return x
            
        model = Model()
        
        data_list = pyg_data_list_from_graphs(graphs)
        assert len(data_list) == num_graphs
        batch = next(iter(DataLoader(data_list, batch_size=num_graphs)))
        out = model(batch)
        
        assert isinstance(out, torch.Tensor)
        assert out.shape == (num_graphs, num_graph_labels)
        

# == JAX/JRAPH ==

class TestJraph:
    
    def test_jraph_from_graph_basically_works(self):
        """
        ``jraph_from_graph`` should take a graph dict as input and turn it into a jraph.GraphsTuple object.
        """
        num_nodes = 15
        num_node_attributes = 20
        num_edge_attributes = 10
        
        graph = create_mock_graph(
            num_nodes=15,
            num_node_attributes=num_node_attributes,
            num_edge_attributes=num_edge_attributes,
        )
        graph_tuple: jraph.GraphsTuple = jraph_from_graph(graph)

        assert isinstance(graph_tuple, jraph.GraphsTuple)
        
        # Node features
        assert int(graph_tuple.n_node[0]) == num_nodes
        assert graph_tuple.nodes.shape == (num_nodes, num_node_attributes)
        
        # Since the mock graphs are just loops, there are as many edges as nodes
        assert int(graph_tuple.n_edge[0]) == num_nodes
        assert graph_tuple.edges.shape == (num_nodes, num_edge_attributes)
        assert graph_tuple.senders.shape == (num_nodes, )
        assert graph_tuple.receivers.shape == (num_nodes, )
        
    def test_jraph_batch_from_graphs_basically_works(self):
        """
        ``jraph_batch_from_graphs`` should create a jraph batched GraphsTuple instance using a list of graph dicts 
        as the input.
        """
        num_graphs = 10
        num_nodes_range = (5, 10)
        num_node_attributes = 20
        num_edge_attributes = 10
        
        graphs = sample_mock_graphs(
            k=num_graphs,
            num_nodes_range=num_nodes_range,
            num_node_attributes=num_node_attributes,
            num_edge_attributes=num_edge_attributes,
        )
        
        # The aggregated batch also is again a GraphsTuple instance
        batch = jraph_implicit_batch_from_graphs(graphs)
        print(batch)
        assert isinstance(batch, jraph.GraphsTuple)
        assert len(jraph.unbatch(batch)) == num_graphs
        
    def test_jraph_model_compatibility(self):
        """
        It should be possible to use the converted jraph graphs to feed them into a simple jraph graph network.
        """
        num_graphs = 15
        num_nodes_range = (5, 10)
        num_node_attributes = 10
        num_edge_attributes = 6
        num_graph_labels = 3
        
        graphs = sample_mock_graphs(
            k=num_graphs,
            num_nodes_range=num_nodes_range,
            num_node_attributes=num_node_attributes,
            num_edge_attributes=num_edge_attributes,
            num_graph_labels=num_graph_labels,
        )
        batch = jraph_implicit_batch_from_graphs(graphs)
        print(batch)
        
        # Here we just want to construct a very simple graph network
        network = jraph.GraphNetwork(
            update_node_fn=lambda node_features, aggregated_sender_edge_features, *args, **kwargs: node_features,
            update_edge_fn=lambda edge_features, *args, **kwargs: edge_features,
            aggregate_edges_for_nodes_fn=jraph.segment_sum,
        )
        network_jit = jax.jit(network)
        updated_batch = network_jit(batch)
        
        # The result of the graph network should again be a graphs tuple instance with the same 
        # internal make-up but with different features.
        assert isinstance(updated_batch, jraph.GraphsTuple)


#!/usr/bin/env python3
"""
Simple debug script for HOPV15 parser without pycomex dependencies.
"""

import os
import tempfile
from chem_mat_data.connectors import FileDownloadSource
from chem_mat_data.data import HOPV15Parser

def main():
    print("Testing HOPV15 Parser on large dataset...")
    print("=" * 50)

    # Download the dataset using the same connector as the original script
    print("Downloading dataset...")
    try:
        with FileDownloadSource(
            'https://figshare.com/ndownloader/files/4513735',
            verbose=True,
            ssl_verify=False,
        ) as source:
            path = source.fetch()
            print(f"Downloaded to: {path}")

            # Check file size
            file_size = os.path.getsize(path)
            print(f"File size: {file_size:,} bytes")

            # Test the parser
            print("\nTesting HOPV15Parser...")
            parser = HOPV15Parser(path=path)

            # Test parse_all
            print("Parsing all molecules...")
            mol_tuples = parser.parse_all()
            print(f"Found {len(mol_tuples)} molecules")

            # Show first few molecules
            print("\nFirst 3 molecules:")
            for i, (mol, info) in enumerate(mol_tuples[:3]):
                print(f"{i+1}. SMILES: {info['smiles']}")
                print(f"   Atoms: {mol.GetNumAtoms()}")
                print(f"   Has exp. properties: {'experimental_properties' in info}")
                if 'experimental_properties' in info:
                    exp_props = info['experimental_properties']
                    print(f"   PCE: {exp_props.get('PCE', 'N/A')}")
                print()

    except Exception as e:
        print(f"Error: {e}")
        import traceback
        traceback.print_exc()

if __name__ == "__main__":
    main()


import io
from typing import List, Dict
import os
import shutil
import gzip
import requests
import pandas as pd

from rdkit import Chem
from pycomex.functional.experiment import Experiment
from pycomex.utils import folder_path, file_namespace


# :param DATASET_NAME:
#       This string determines the name of the message pack dataset file that is then 
#       stored into the "results" folder of the experiment as the result of the 
#       processing process. The corresponding file extensions will be added 
#       automatically.
DATASET_NAME: str = 'compas_3x'
# :param METADATA:
#       A dictionary which will be used as the basis for the metadata that will be added 
#       as additional information to the file share server.
METADATA: dict = {
    'description': (
        'The third installment of the COMPAS Project, a computational database of polycyclic aromatic systems, '
        'focused on peri-condensed polybenzenoid hydrocarbons. This dataset contains optimized ground-state structures '
        'and a selection of molecular properties for approximately 39k and 9k peri-condensed polybenzenoid hydrocarbons '
        'at different computational levels. The dataset supports data-driven analysis of structure-property trends and '
        'is useful for machine- and deep-learning studies in chemistry.'
    ),
    'target_type': ['Regression'],
    'tags': ['SMILES', 'Molecules', 'Quantum Chemistry', 'Molecular Properties'],
    'sources': [
        'https://pubs.rsc.org/en/content/articlelanding/2024/cp/d4cp01027b',
        'https://gitlab.com/porannegroup/compas/-/tree/main/COMPAS-3?ref_type=heads',
    ],
    'target_descriptions': {
        0: 'HOMO_eV - energy of the highest molecular orbit (HOMO) in electron volt (eV)',
        1: 'LUMO_eV - energy of the lowest unoccupied molecular orbit (LUMO) in electron volt (eV)',
        2: 'GAP_eV - energy gap between HOMO and LUMO in electron volt (eV)',
        3: 'Dipmom_Debye - dipole moment in Debye',
        4: 'Etot_eV - total energy of the molecule in electron volt (eV)',
        5: 'aEA_eV - adiabatic electron affinity in electron volt (eV)',
        6: 'aIP_eV - adiabatic ionization potential in electron volt (eV)',
        7: 'NFOD - fractional occupation density',
        8: 'n_rings - number of rings in the molecule',
    }
}

__TESTING__ = False

experiment = Experiment.extend(
    'create_graph_datasets.py',
    base_path=folder_path(__file__),
    namespace=file_namespace(__file__),
    glob=globals(),
)


@experiment.hook('add_graph_metadata', default=False, replace=True)
def add_graph_metadata(e: Experiment, data: dict, graph: dict) -> dict:
    """
    We add the compound id for identification and the molecular weight
    """
    pass #graph['graph_id'] = data['ID']


@experiment.hook('load_dataset', default=False, replace=True)
def load_dataset(e: Experiment) -> Dict[int, dict]:
    
    # For the COMPAS 2 dataset the raw file is provided in the form of a CSV file that is hosted on Gitlab
    # here we download that file and then open the content as a pandas dataframe.
    e.log('downloading raw csv file...')
    raw_url = 'https://gitlab.com/porannegroup/compas/-/raw/main/COMPAS-3/compas-3x.csv?ref_type=heads&inline=false'
    response = requests.get(raw_url, verify=False)
    response.raise_for_status()  # Ensure the request was successful
    csv_file = io.StringIO(response.text)
    df = pd.read_csv(csv_file, quotechar='"')
    
    print(df.head())
    
    compressed_path = os.path.join(e.path, f'{e.DATASET_NAME}.csv.gz')
    with gzip.open(compressed_path, mode='wb') as compressed_file:
        shutil.copyfileobj(csv_file, compressed_file)

    columns = [
        'HOMO_eV',
        'LUMO_eV',
        'GAP_eV',
        'Dipmom_Debye',
        'Etot_eV',
        'aEA_eV',
        'aIP_eV',
        'NFOD',
        'n_rings',
    ]

    dataset: Dict[int, dict] = {}
    index: int = 0
    for c, data in enumerate(df.to_dict('records')):
        data['smiles'] = data['smiles']
        
        # == MOLECULE FILTERS ==
        # We don't want to use compounds with '.' in the smiles (separate molecules)
        if '.' in data['smiles']:
            continue
        
        # We don't want to use compounds that only consist of a single atom
        mol = Chem.MolFromSmiles(data['smiles'])
        if not mol:
            continue
        
        # We also don't want to accept "molecules" that are essentially just individual atoms
        if len(mol.GetAtoms()) < 2:
            continue
        
        # == TARGETS ==
        # In this dataset we only have one target
        data['targets'] = [data[key] for key in columns]
        dataset[index] = data
        
        index += 1
        
        if c % 10_000 == 0:
            print(f' * processed {c} records')

    return dataset

experiment.run_if_main()


"""
This module tests all the datasets that are availabile on remote file share server 
by obtaining each dataset and attempting to use them for a simple GNN forward pass.
"""
import numpy as np
import pytest
from typing import Dict, List

from chem_mat_data.testing import ConfigIsolation
from chem_mat_data.testing import assert_graph_dict
from chem_mat_data.web import AbstractFileShare
from chem_mat_data.web import construct_file_share
from chem_mat_data.main import load_graph_dataset
from chem_mat_data.main import pyg_data_list_from_graphs


def fetch_datasets():
    
    with ConfigIsolation() as iso:
        
        file_share_type = iso.config.get_fileshare_type()
        file_share: AbstractFileShare = construct_file_share(
            file_share_type=iso.config.get_fileshare_type(),
            file_share_url=iso.config.get_fileshare_url(),
            file_share_kwargs=iso.config.get_fileshare_parameters(file_share_type),    
        )
        metadata: dict = file_share.fetch_metadata()
        
        datasets: Dict[str, dict] = metadata['datasets']
        
        # Now we also want to be prudent and filter datasets that are too large so we don't significantly slow 
        # down the testing.
        
        dataset_names: List[str] = []
        for dataset_name, dataset_info in datasets.items():

            # Skip datasets that are too large and would take too long to process
            if dataset_info['compounds'] >= 200_000:
                continue

            # We also want to skip the hidden datasets that start with an underscore
            if dataset_name.startswith('_'):
                continue

            # Only test datasets that have full graph data available (full: true)
            if not dataset_info.get('full', False):
                continue

            dataset_names.append(dataset_name)
            
        # Finally we return the list of the dataset names to be used as the basis of the individual test
        # cases.
        return dataset_names
    
    
# Fetch datasets 
@pytest.mark.parametrize('dataset_name', fetch_datasets())
def test_dataset_torch_forward_pass(dataset_name: str):
    
    with ConfigIsolation() as iso:
        
        file_share_type = iso.config.get_fileshare_type()
        file_share: AbstractFileShare = construct_file_share(
            file_share_type=iso.config.get_fileshare_type(),
            file_share_url=iso.config.get_fileshare_url(),
            file_share_kwargs=iso.config.get_fileshare_parameters(file_share_type),    
        )
        metadata: dict = file_share.fetch_metadata()
        dataset_info = metadata['datasets'][dataset_name]
        
        # ~ loading the graphs from the remote file share
        graphs: List[dict] = load_graph_dataset(
            dataset_name=dataset_name,
            folder_path=iso.config.cache_path,
            config=iso.config,
        )
        assert len(graphs) != 0
        assert len(graphs) == dataset_info['compounds'], (
            f'The number of graphs in the dataset ({len(graphs)}) should match the number of compounds reported '
            f'in the metadata file ({dataset_info["compounds"]}) for dataset {dataset_name}'
        )
        example_graph = graphs[0]
        
        # ~ checking the individual graphs for validity
        for index, graph in enumerate(graphs):
            assert_graph_dict(graph)
            
        # ~ Doing model forward pass
        try:
            from torch import Tensor
            from torch_geometric.data import Data  # noqa: F401
            from torch_geometric.loader import DataLoader
            from torch_geometric.nn.models import GIN

            data_list = pyg_data_list_from_graphs(graphs)
            data_loader = DataLoader(data_list, batch_size=128, shuffle=True)
            
            model = GIN(
                in_channels=example_graph['node_attributes'].shape[1],
                out_channels=example_graph['graph_labels'].shape[0],
                hidden_channels=64,
                num_layers=3,
            )
            for data in data_loader:
                out_pred: Tensor = model.forward(
                    x=data.x,
                    edge_index=data.edge_index,
                )
                out_pred: np.ndarray = out_pred.cpu().detach().numpy()
                assert not (np.isnan(out_pred).any() or np.isinf(out_pred).any())
            
        except (ImportError, ModuleNotFoundError):
            
            print('Skipping forward pass test because PyTorch is not installed...')
            
        
        


"""
from typing import List, Dict
This experiment downloads the MOLT-4 dataset from the TU dataset collection which is a classification 
dataset related to  Leukemia. 
The dataset consists of molecular graphs and binary class labels.
"""
from typing import Optional, Dict

import rdkit.Chem as Chem

from pycomex.functional.experiment import Experiment
from pycomex.utils import folder_path, file_namespace

# :param SOURCE_URL:
#       As a fallback option, if no local dataset path is given, this parameter can define the URL at 
#       which the dataset can be downloaded from. This should be a valid URL that points to the TU dataset 
#       format. If given, the experiment will attempt to download and extract before loading the dataset.
SOURCE_URL: Optional[str] = 'https://www.chrsmrrs.com/graphkerneldatasets/MOLT-4.zip'
# :param DATASET_NAME:
#       The name of the dataset which will be used as the final file name aka the unique string 
#       identifier on the remote file share server.
DATASET_NAME: str = 'MOLT-4'
# :param NODE_LABEL_MAP:
#       This is a dictionary that maps the node label classes to the integer values corresponding to 
#       the actual atomic numbers. This mapping will have to be extracted from the README file of 
#       the TU dataset.
NODE_LABEL_MAP: Dict[int, str] = {
    0: 'O',
    1: 'N',
    2: 'C',
    3: 'Br',
    4: 'S',
    5: 'Cl',
    6: 'F',
    7: 'Na',
    8: 'Sn',
    9: 'Pt',
    10: 'Ni',
    11: 'Zn',
    12: 'Mn',
    13: 'P',
    14: 'I',
    15: 'Cu',
    16: 'Co',
    17: 'Se',
    18: 'Au',
    19: 'Ge',
    20: 'Fe',
    21: 'Pb',
    22: 'Si',
    23: 'B',
    24: 'Nd',
    25: 'In',
    26: 'Bi',
    27: 'Er',
    28: 'Hg',
    29: 'As',
    30: 'Ga',
    31: 'Ti',
    32: 'Ac',
    33: 'Y',
    34: 'Eu',
    35: 'Tl',
    36: 'Zr',
    37: 'Hf',
    38: 'K',
    39: 'La',
    40: 'Ce',
    41: 'Sm',
    42: 'Gd',
    43: 'Dy',
    44: 'U',
    45: 'Pd',
    46: 'Ir',
    47: 'Re',
    48: 'Li',
    49: 'Sb',
    50: 'W',
    51: 'Mg',
    52: 'Ru',
    53: 'Rh',
    54: 'Os',
    55: 'Th',
    56: 'Mo',
    57: 'Nb',
    58: 'Ta',
    59: 'Ag',
    60: 'Cd',
    61: 'V',
    62: 'Te',
    63: 'Al',
}
# :param EDGE_LABEL_MAP:
#       This is a dictionary that maps the edge label classes to the integer values corresponding to
#       the actual bond types. This mapping will have to be extracted from the README file of
#       the TU dataset.
EDGE_LABEL_MAP: Dict[str, int] = {
    0: Chem.BondType.SINGLE,
    1: Chem.BondType.DOUBLE,
    2: Chem.BondType.TRIPLE,
}
# :param GRAPH_LABEL_MAP:
#       This is a dictionary that maps the graph label classes to the integer values corresponding to
#       the actual graph labels. This mapping will have to be extracted from the README file of
#       the TU dataset.
GRAPH_LABEL_MAP: Dict[int, list] = {
    1: [0, 1],
    0: [1, 0],
}

experiment = Experiment.extend(
    'create_graph_datasets__from_tu.py',
    base_path=folder_path(__file__),
    namespace=file_namespace(__file__),
    glob=globals(),
)

experiment.run_if_main()




from typing import List, Dict
from pycomex.functional.experiment import Experiment
from pycomex.utils import folder_path, file_namespace

from chem_mat_data.data import QM9XyzParser

# == SOURCE PARAMETERS ==

# :param SOURCE_PATH:
#       This is the absolute string path of the dataset in xyz format. This can either be an archive or 
#       directly a folder which points to an "xyz bundle". This bundle has to ultimately contain a number 
#       of distinct .XYZ files where each file defines one element (molecule) in the target dataset.
#       Optionally, this xyz bundle may also contain a csv file with additional information about each 
#       of the elements in the dataset.
SOURCE_PATH: str = '/home/jonas/Downloads/qm9.xyz_bundle'
# :param PARSER_CLASS:
#       This has to define the class that is to be used for the parsing of the xyz files. The class that 
#       is defined here needs to be a valid subclass of the AbstractXyzParser interface. There are different 
#       flavors of xyz files for which different parser classes exist.
PARSER_CLASS: type = QM9XyzParser

# == PROCESSING PARAMETERS ==
# These parameters can be used to configure the processing functionality of the script 
# itself. This includes for example whether or not the molecule coordinates should be 
# created by RDKIT as well or not. Also the compression of the final dataset file can
# be configured here.

# :param DATASET_NAME:
#       This string determines the name of the message pack dataset file that is then 
#       stored into the "results" folder of the experiment as the result of the 
#       processing process. The corresponding file extensions will be added 
#       automatically.
DATASET_NAME: str = 'qm9'

experiment = Experiment.extend(
    'create_graph_datasets__from_xyz.py',
    base_path=folder_path(__file__),
    namespace=file_namespace(__file__),
    glob=globals(),
)

experiment.run_if_main()


import os
from typing import List

import lorem
import matplotlib.pyplot as plt
import chem_mat_data._typing as tc
import rdkit.Chem as Chem
from rich.console import Console
from rich.pretty import pprint
from chem_mat_data.testing import assert_graph_dict
#from chem_mat_data.graph import assert_graph_dict
from chem_mat_data.processing import OneHotEncoder
from chem_mat_data.processing import RichProcessingSummary
from chem_mat_data.processing import CrippenEncoder
from chem_mat_data.processing import MoleculeProcessing
from chem_mat_data.data import load_xyz_as_mol

from .utils import ARTIFACTS_PATH, ASSETS_PATH


class TestMoleculeProcessing:
    
    DEFAULT_SMILES = 'C1=CC=CC=C1CCO'
    
    def test_basically_works(self):
        """
        Constructs a new instance of the class. "process" method should turn a SMILES into a 
        GraphDict representation
        """
        processing = MoleculeProcessing()
        assert isinstance(processing, MoleculeProcessing)
        
        smiles = self.DEFAULT_SMILES
        graph: tc.GraphDict = processing.process(smiles)
        
        # This function will run a series of assert statements that checks the validity of the graph dict
        # formatting. So if it contains the correct keys and the correct array shapes etc.
        assert_graph_dict(graph)
        pprint(graph)
        
    def test_process_includes_node_atom_symbols(self):
        """
        When processing a smiles string the resulting graph dict should also contian the "node_atoms" entry which 
        is essentially a list of string representations of the atom symbols.
        """
        processing = MoleculeProcessing()
        
        smiles = self.DEFAULT_SMILES
        graph: tc.GraphDict = processing.process(smiles)
        
        assert 'node_atoms' in graph
        print(graph['node_atoms'])
        assert len(graph['node_indices']) == len(graph['node_atoms'])
        assert isinstance(graph['node_atoms'][0], str)
        
    def test_process_includes_edge_bond_types(self):
        """
        When processing a smiles string the resulting graph dict should also contain the "edge_bonds" entry which
        is essentially a list of string representations of the bond types.
        """
        processing = MoleculeProcessing()
        
        smiles = self.DEFAULT_SMILES
        graph: tc.GraphDict = processing.process(smiles)
        
        assert 'edge_bonds' in graph
        print(graph['edge_bonds'])
        assert len(graph['edge_bonds']) == len(graph['edge_indices'])
        assert isinstance(graph['edge_bonds'][0], str)
        
    def test_process_includes_graph_representation(self):
        """
        When processing a smiles string the resulting graph dict should also contain the "graph_repr" entry which 
        is essentially a string representation of the original SMILES string.
        """
        processing = MoleculeProcessing()
        
        smiles = self.DEFAULT_SMILES
        graph: tc.GraphDict = processing.process(smiles)
        
        assert 'graph_repr' in graph
        assert isinstance(graph['graph_repr'], str)
        assert graph['graph_repr'] == smiles
        
    def test_visualize_as_figure(self):
        """
        visualize_as_figure should use the rdkit SVG engine to draw the given molecule and 
        and project it into a matplotlib Figure which we can then save to the disk as PNG
        """
        processing = MoleculeProcessing()
        smiles = self.DEFAULT_SMILES
        fig, _ = processing.visualize_as_figure(smiles, width=1000, height=1000)

        assert isinstance(fig, plt.Figure)
        
        fig_path = os.path.join(ARTIFACTS_PATH, 'test_molecule_processing_visualize_as_figure.png')
        fig.savefig(fig_path)
        assert os.path.exists(fig_path)
        
    def test_summary(self):
        """
        The summary method should return / print a summary and description of the molecule processing 
        steps aka descriptions of the node and edge attributes that were extracted for the molecular 
        graph.
        """
        processing = MoleculeProcessing()
        summary = processing.summary(echo=False)
        st = str(summary)
        print('\n', st)
        
        assert isinstance(st, str)
        assert len(st) != 0

    def test_process_mol_object_works(self):
        """
        In principle, the process function is designed to work on the SMILES representation of a molecule 
        but it should also work with the rdkit Mol object itself as converting to a Mol is the first 
        thing that happens in the process function anyway.
        """
        smiles = self.DEFAULT_SMILES
        mol = Chem.MolFromSmiles(smiles)
        
        processing = MoleculeProcessing()
        graph = processing.process(mol)
        assert isinstance(graph, dict)
        assert_graph_dict(graph)
        
    def test_process_mol_object_from_xyz_file_works(self):
        """
        As a special case of processing a RDKit molecule object, we can also load a molecule from a
        xyz file and then process it into a graph representation. This is perhaps a bit different 
        because by default the Mol object that is loaded from the xyz file does not contain any 
        bonds/edges.
        """
        xyz_path = os.path.join(ASSETS_PATH, '_test.xyz')
        mol, info = load_xyz_as_mol(xyz_path)
        print(mol, mol.GetNumAtoms())
        assert isinstance(mol, Chem.Mol)
        assert isinstance(info, dict)
        
        processing = MoleculeProcessing()
        graph = processing.process(mol)
        assert isinstance(graph, dict)
        assert_graph_dict(graph)
        
        # Another specialty of processing a mol from an xyz file is that the graph representation
        # should also contain the 3D node coordinates.
        assert 'node_coordinates' in graph


class TestCrippenEncoder:
    
    DEFAULT_SMILES = 'C1=CC=CC=C1CCO'
    
    def test_basically_works(self):
        
        encoder = CrippenEncoder()
        
        mol = Chem.MolFromSmiles(self.DEFAULT_SMILES)
        for atom in mol.GetAtoms():
            value: List[float] = encoder.encode(mol, atom)
            print(value)
            
            assert isinstance(value, list)
            assert len(value) > 0
            # We know that the crippen contribution consists of two values.
            assert len(value) == 2 

    def test_description_works(self):
        """
        The descriptions property should return a list of strings which describe the encoding of 
        the two values that are created by the encoder.
        """
        encoder = CrippenEncoder()
        descriptions = encoder.descriptions
        
        assert isinstance(descriptions, list)
        assert len(descriptions) == 2
        for desc in descriptions:
            assert isinstance(desc, str)
            assert len(desc) > 0
            

class TestRichProcessingSummary:
    
    DEFAULT_SECTIONS = {
            'section 1': {
                0: lorem.sentence(),
                1: lorem.paragraph(),
            },
            'section 2': {
                0: lorem.sentence(),
                1: lorem.sentence(),
                2: lorem.sentence(), 
            }
        }
    
    def test_basically_works(self):
            
        rich_summary = RichProcessingSummary(self.DEFAULT_SECTIONS)
        assert isinstance(rich_summary, RichProcessingSummary)
        
        console = Console()
        with console.capture() as capture:
            console.print(rich_summary)
            
        st = capture.get()
        print('')
        print(st)
        
        # Checking for some sections whether they are in the string
        assert 'section 1' in st
        assert 'section 2' in st
        assert '00' in st
        
    def test_str_conversion_works(self):
        """
        RichProcessingSummary extends RichMixin, which provides a default implementation for 
        the __str__ method. Therefore it should be possible to call str() on the object and 
        get a plain text representation of the string.
        """
        rich_summary = RichProcessingSummary(self.DEFAULT_SECTIONS)
        st = str(rich_summary)
        print(st)
        
        assert isinstance(st, str)
        assert len(st) > 0
        
        # Checking for some sections whether they are in the string
        assert 'section 1' in st
        assert 'section 2' in st
        assert '00' in st


class TestOneHotEncoder:
    
    def test_basically_works(self):
        """
        The OneHotEncoder should be able to encode and decode values from a list of categories.
        """
        encoder = OneHotEncoder(['a', 'b', 'c'], add_unknown=False, dtype=str)
        
        # the most simple test cases
        assert encoder.encode('a') == [1., 0., 0.]
        assert encoder.encode('b') == [0., 1., 0.]
        assert encoder.encode('c') == [0., 0., 1.]
        
        # Without an unknown this should simply not be encoded at all
        assert encoder.encode('x') == [0., 0., 0.]
        
        # decoding should also work for the basic cases
        assert encoder.decode([1., 0., 0.]) == 'a'
        assert encoder.decode([0., 1., 0.]) == 'b'
        assert encoder.decode([0., 0., 1.]) == 'c'
        
    def test_unknown_works(self):
        """
        When construction the encoder, we can specify that we want to add an "unknown" category
        in which case all values not in the original list will be encoded into a additional 
        category "unknown".
        """
        encoder = OneHotEncoder(
            ['a', 'b', 'c'], 
            add_unknown=True, 
            dtype=str,
            unknown='?'    
        )
        
        assert encoder.encode('a') == [1., 0., 0., 0.]
        assert encoder.encode('x') == [0., 0., 0., 1.]
        
        # Also the decoding operation should now return the default value for the unkown category
        # when decoding such an unknown vector.
        assert encoder.decode([0., 0., 0., 1.]) == '?'
        
    def test_descriptions_basically_work(self):
        """
        The "descriptions" property should return a list of strings which describe the encoding.
        """
        encoder = OneHotEncoder(['a', 'b', 'c'], add_unknown=False, dtype=str)
        
        # "descriptions" is a property which should create a list of string descriptions
        descriptions = encoder.descriptions
        assert isinstance(descriptions, list)
        assert descriptions == ['is a?', 'is b?', 'is c?']
        
    def test_custom_descriptions_work(self):
        """
        It should be possible to provide custom descriptions for the encoder categories in the 
        constructor. These descriptions will then be used instead of the actual values to generate 
        the descriptions list.
        """
        encoder = OneHotEncoder(
            ['a', 'b', 'c'], 
            add_unknown=False, 
            dtype=str,
            value_descriptions=['letter A', 'letter B', 'letter C']
        )
        
        descriptions = encoder.descriptions
        assert descriptions == ['is letter A?', 'is letter B?', 'is letter C?']


"""
from typing import List, Dict
This experiment module is the base module that handles the creation of a graph dataset from a 
raw representation as an "xyz bundle". In the most basic form, such an xyz bundle represents a target 
dataset as a folder or an archive consisting of multiple .xyz files where each file represents one 
element of the dataset. These .xyz files already contain additional information about the 3D coordinates 
of the atoms.
"""
import os
import csv
import pandas as pd
from typing import Dict, List

import numpy as np
from pycomex.functional.experiment import Experiment
from pycomex.utils import folder_path, file_namespace

from chem_mat_data.main import load_xyz_dataset
from chem_mat_data.utils import SCRIPTS_PATH
from chem_mat_data.data import QM9XyzParser

# :param SOURCE_PATH:
#       This is the absolute string path of the dataset in xyz format. This can either be an archive or 
#       directly a folder which points to an "xyz bundle". This bundle has to ultimately contain a number 
#       of distinct .XYZ files where each file defines one element (molecule) in the target dataset.
#       Optionally, this xyz bundle may also contain a csv file with additional information about each 
#       of the elements in the dataset.
SOURCE_PATH: str = os.path.join(SCRIPTS_PATH, 'assets', '_test.xyz_bundle')
# :param PARSER_CLASS:
#       This has to define the class that is to be used for the parsing of the xyz files. The class that 
#       is defined here needs to be a valid subclass of the AbstractXyzParser interface. There are different 
#       flavors of xyz files for which different parser classes exist.
PARSER_CLASS: type = QM9XyzParser
# :param DATASET_NAME:
#       The name of the dataset which will be used as the final file name aka the unique string 
#       identifier on the remote file share server.
DATASET_NAME: str = 'qm9'

__TESTING__ = False

experiment = Experiment.extend(
    'create_graph_datasets.py',
    base_path=folder_path(__file__),
    namespace=file_namespace(__file__),
    glob=globals(),
)


@experiment.hook('add_graph_metadata', default=False, replace=True)
def add_graph_metadata(e: Experiment, data: dict, graph: dict) -> dict:
    """
    No extra metadata
    """
    pass


@experiment.hook('load_dataset', default=False, replace=True)
def load_dataset(e: Experiment) -> Dict[int, dict]:
    """
    """
    e.log('loading dataset from "xyz_bundle"...')
    
    if os.path.exists(e.SOURCE_PATH):
        e.log(f'the given source path exists @ {e.SOURCE_PATH}')
        source_file = os.path.basename(e.SOURCE_PATH)
        source_folder = os.path.dirname(e.SOURCE_PATH)
        
        df: pd.DataFrame = load_xyz_dataset(
            dataset_name=source_file,
            folder_path=source_folder,
            parser_cls=e.PARSER_CLASS,
            use_cache=True,
        )
    
    else:
        e.log('the given source path does not exist on the local system!')
        e.log(f'attempting to fetch from the remote file share as "{e.SOURCE_PATH}"...')
        
        df: pd.DataFrame = load_xyz_dataset(
            dataset_name=e.SOURCE_PATH,
            parser_cls=e.PARSER_CLASS,
            use_cache=True,
        )
        
    index_data_map: Dict[int, dict] = {}
    for idx, row in df.iterrows():
        index_data_map[idx] = row.to_dict()
    
    return index_data_map


@experiment.hook('after_save', default=False, replace=False)
def after_save(e: Experiment,
               index_data_map: Dict[int, dict],
               graphs: List[dict],
               **kwargs,
               ) -> None:
    """
    This hook is called after the graphs dataset itself was already saved.
    
    ---
    
    In the case of the xyz bundle datasets, we will create a meta.csv file which contains all the 
    additional information about the dataset.
    """
    e.log('creating meta.csv file...')
    
    # ~ figuring out the keys
    # First of all we need to figure out which of the keys in the "data" dicts we can actually 
    # reasonably export into a csv file. The problem is that this dict also contains the "mol"
    # key which for example is a Chem.Mol instance and therefore not able to export into the 
    # csv format.
    # To do this we loop over all the keys of an example element and check the types.
    example_data: dict = next(iter(index_data_map.values()))
    keys: List[str] = []
    for key, value in example_data.items():
        
        # if the value is a primitive data type we use it for sure.
        if isinstance(value, (str, int, float)):
            keys.append(key)

        # if the value is a one-dimensional xyz file then we also include it.                
        if isinstance(value, (np.ndarray)):
            if value.ndim <= 1:
                keys.append(key)
                
    e.log(f'using keys: {keys}')

    # ~ saving the csv file.
    # Now we can simply iterate over all the elements in our dataset and save the dataset
    
    csv_path = os.path.join(e.path, 'meta.csv')
    with open(csv_path, mode='w') as file:
        dict_writer = csv.DictWriter(file, fieldnames=keys)
        dict_writer.writeheader()
        for index, data in index_data_map.items():
            dict_writer.writerow({key: data[key] for key in keys})
    
    e.log(f'saved meta.csv file @ {csv_path}')


experiment.run_if_main()



import pandas as pd
import rdkit.Chem as Chem
from typing import List, Dict
from rich.pretty import pprint
from pycomex.functional.experiment import Experiment
from pycomex.utils import folder_path, file_namespace

from chem_mat_data import load_smiles_dataset
from chem_mat_data.connectors import FileDownloadSource
from chem_mat_data.data import HOPV15Parser

DATASET_NAME: str = 'hopv15_exp'
DESCRIPTION: str = (
    'The Harvard Organic Photovoltaic Dataset (HOPV15) is a comprehensive collection of '
    'experimental photovoltaic data from the literature for 350 different organic solar '
    'cell donor structures, combined with corresponding quantum-chemical calculations '
    'performed over multiple molecular conformers. The dataset includes both experimental '
    'measurements (such as power conversion efficiency, HOMO/LUMO energies, and other '
    'photovoltaic characteristics) and computational results using various density '
    'functionals and basis sets. This resource is designed to enable calibration of '
    'quantum chemical calculations to experimental observations, support the development '
    'of new computational methodologies, and benchmark current and future model chemistries '
    'for organic electronic applications.'
)

# :param METADATA:
#       A dictionary which will be used as the basis for the metadata that will be added 
#       as additional information to the file share server.
METADATA: dict = {
    'tags': [
        'Molecules', 
        'SMILES', 
        'Organic Electronics',
        'Solar Cells',
    ],
    'verbose': 'Harvard Organic Photovoltaic Dataset',
    'sources': [
        'https://chemrxiv.org/engage/chemrxiv/article-details/64be471cb605c6803b425da6'
    ],
    # TADF: Thermally Activated Delayed Fluorescence
    'target_descriptions': {
        '0': 'homo - Highest occupied molecular orbit in eV',
        '1': 'lumo - Lowest unoccupied molecular orbit in eV',
        '2': 'gap - Gap between HOMO and LUMO energy levels in eV',
        '3': 'JSC - Closed circuit current in mA/cm2',
        '4': 'VOC - Open circuit voltage in V',
        '5': 'PCE - Power conversion efficience in percent'
    }
}

experiment = Experiment.extend(
    'create_graph_datasets.py',
    base_path=folder_path(__file__),
    namespace=file_namespace(__file__),
    glob=globals(),
)

@experiment.hook('load_dataset', default=False, replace=True)
def load_dataset(e: Experiment) -> Dict[int, dict]:
    
    # This file contains the dataset in question. It is available as a download link from a 
    # Chemarxiv paper.
    e.log('Downloading dataset...')
    with FileDownloadSource(
        'https://figshare.com/ndownloader/files/4513735',
        verbose=True,
        ssl_verify=False,
    ) as source:
        path = source.fetch()
        
        parser = HOPV15Parser(path=path)
        mol_tuples = parser.parse_all()
        e.log(f'Downloaded dataset with {len(mol_tuples)} entries')
        
        pprint(mol_tuples[:3], max_depth=3)
        pprint(mol_tuples[0][1]['experimental_properties'], max_depth=3)
        
        # Convert the mol_tuples into a dataframe format
        df_data = []
        for mol, data in mol_tuples:
            
            smiles = data.get('smiles')
            # Create a row with SMILES and experimental properties
            if data.get('experimental_properties'):
                row = {
                    'smiles': smiles,
                    **data['experimental_properties']  # Flatten experimental properties into the row
                }
                df_data.append(row)
            else:
                e.log(f'No experimental properties for SMILES: {smiles}')

        df = pd.DataFrame(df_data)
        print(df.head())
        e.log(f'Created dataframe with {len(df)} rows and columns: {list(df.columns)}')
        
    
    # Now that we have the raw data as a dataframe, we can convert it into the expected 
    # dictionary format where each entry in the dictionary corresponds to a single data point 
    # in the dataset.
    e.log('Processing dataset...')
    dataset: Dict[int, dict] = {}
    for index, data in enumerate(df.to_dict('records')):

        data['smiles'] = data['smiles']
        
        # == MOLECULE FILTERS ==
        # We don't want to use compounds with '.' in the smiles (separate molecules)
        if '.' in data['smiles']:
            continue
        
        # We don't want to use compounds that only consist of a single atom
        mol = Chem.MolFromSmiles(data['smiles'])
        if not mol:
            continue
        
        # We also don't want to accept "molecules" that are essentially just individual atoms
        if len(mol.GetAtoms()) < 2:
            continue
        
        # == TARGET VALUES ==
        # Only one target. We combine the experimental and computational data into a single list 
        data['targets'] = [
            data['HOMO'],
            data['LUMO'],
            data['gap'],
            data['JSC'],
            data['VOC'],
            data['PCE'],
        ]

        dataset[index] = data

    return dataset

experiment.run_if_main()




from chem_mat_data.utils import get_version


# A function name needs to start with "test_" for pytest to be able to automatically 
# discover it.
def test_get_version():
    version = get_version()
    
    # With pytest - unlike unittest - it is sufficient to use the standard "assert" syntax.
    # the second string argument to the assert statement is the error message that will be 
    # printed in case the assertion fails.
    assert isinstance(version, str), 'the version is not a string!'
    assert version != '', 'the version is an empty string'
    




"""
This module tests the unittest utilities that in the "testing" module
"""
import os
import numpy as np

from chem_mat_data.config import Config
from chem_mat_data.cache import Cache
from chem_mat_data.testing import create_mock_graph
from chem_mat_data.testing import sample_mock_graphs
from chem_mat_data.testing import ConfigIsolation


def test_create_mock_graph_basically_works():
    """
    ``create_mock_graph`` should create a valid graph dict representation.
    """
    
    num_nodes = 20
    num_node_attributes = 10
    num_edge_attributes = 8
    num_graph_labels = 1
    
    graph = create_mock_graph(
        num_nodes=num_nodes,
        num_node_attributes=num_node_attributes,
        num_edge_attributes=num_edge_attributes,
        num_graph_labels=num_graph_labels
    )
    assert isinstance(graph['node_indices'], np.ndarray)
    assert graph['node_indices'].shape == (num_nodes, )
    
    assert isinstance(graph['node_attributes'], np.ndarray)
    assert graph['node_attributes'].shape == (num_nodes, num_node_attributes)
    
    assert isinstance(graph['edge_indices'], np.ndarray)
    assert graph['edge_indices'].shape == (num_nodes, 2)
    
    assert isinstance(graph['edge_attributes'], np.ndarray)
    assert graph['edge_attributes'].shape == (num_nodes, num_edge_attributes)
    
    assert isinstance(graph['node_indices'], np.ndarray)
    assert graph['graph_labels'].shape == (num_graph_labels, )
    
    
def test_sample_mock_graphs_basically_works():
    """
    ``sample_mock_graphs`` should return a number of mock graphs with different numbers of 
    nodes randomly sampled in a certain range.
    """
    num_graphs = 50
    num_nodes_min = 5
    num_nodes_max = 30
    
    graphs = sample_mock_graphs(
        k=num_graphs,
        num_nodes_range=(num_nodes_min, num_nodes_max)
    )
    
    assert isinstance(graphs, list)
    assert len(graphs) == num_graphs
    
    for graph in graphs:
        assert isinstance(graph, dict)
        num_nodes = len(graph['node_indices'])
        assert num_nodes_min <= num_nodes <= num_nodes_max
        
        
def test_config_isolation_basically_works():
    """
    The ConfigIsolation context manager should create a temporary isolated config and cache folder and 
    temporarily overwrite the values of the Config singleton such that there are no side effects during 
    a test case that relies on the Config object state.
    """
    config = Config()
    cache_path = config.cache_path
    config_path = config.config_path
    
    with ConfigIsolation() as iso:
        
        assert config.cache_path != cache_path
        assert config.config_path != config_path
        
        # Also, the Cache instance should exist and should point to the same cache folder
        assert isinstance(config.cache, Cache)
        assert config.cache.path == iso.cache_path
        
        # Also, the config file should exist in the newly created temp folder
        assert os.path.exists(config.config_file_path)
        
    assert config.cache_path == cache_path
    assert config.config_path == config_path


from typing import Dict
from typing import List, Dict
from rdkit import Chem
import pandas as pd
from pycomex.functional.experiment import Experiment
from pycomex.utils import folder_path, file_namespace

import numpy as np

DATASET_NAME: str = 'synth_binary_local'

__TESTING__ = False
__DEBUG__ = True

experiment = Experiment.extend(
    'create_graph_datasets.py',
    base_path=folder_path(__file__),
    namespace=file_namespace(__file__),
    glob=globals(),
)

# This dictionary defines all the relevant SMARTS patterns that we want to use later on 
# to check for local patterns. The keys of this dict are human-readable names for the patterns 
# and the values are the actual SMARTS strings.
SMARTS_PATTERNS: Dict[str, str] = {
    'benzene'           : 'c1ccccc1',
    'nitro'             : '[NX3](=O)=O',
    'halogen'           : '[F,Cl,Br,I]',
    'carboxylic_acid'   : 'C(=O)[OH]',
    'tertiary_amine'    : '[$([NX3]([#6])([#6])[#6])]',  # N with three C neighbors
    'sulfoxide'         : '[#16](=O)',
    'ether'             : 'C-O-C',
    'pyridine'          : 'n1ccccc1',
    'furan'             : 'o1cccc1',
    'ketone'            : 'C(=O)[#6]',
    'alcohol'           : '[OX2H]',
    'thiol'             : '[SX2H]'
}

# compile the SMARTS patterns
SMARTS_PATTERNS_COMPILED = {
    name: Chem.MolFromSmarts(smarts)
    for name, smarts in SMARTS_PATTERNS.items()
}


def extract_motifs(mol: Chem.Mol, smarts_patterns: dict = SMARTS_PATTERNS_COMPILED) -> dict:
    """Return a dict mapping motif name → bool for whether
       that SMARTS appears in the molecule."""
    return {
        name: bool(mol.HasSubstructMatch(pat))
        for name, pat in smarts_patterns.items()
    }


def motif_label(mol: Chem.Mol, smarts_patterns: dict) -> int:
    """
    A rather complicated rule:
    
      Label = 1 if any of these three big clauses holds:
      
        Clause 1:
          benzene AND (nitro OR (halogen AND carboxylic_acid))
        
        Clause 2:
          tertiary_amine AND NOT sulfoxide
          AND (ether OR halogen)
        
        Clause 3:
          (pyridine OR furan) AND ketone AND NOT alcohol
        
      Otherwise 0.
    """
    m = extract_motifs(mol, smarts_patterns=smarts_patterns)
    
    # These are the individual clauses that will suffice to determine the label of the molecule
    # if any of these claususes are true, then the label is 1 otherwise 0.
    c1 = m['benzene'] and (m['nitro'] or (m['halogen'] and m['carboxylic_acid']))
    c2 = m['tertiary_amine'] and not m['sulfoxide'] and (m['ether'] or m['halogen'])
    c3 = (m['pyridine'] or m['furan']) and m['ketone'] and not m['alcohol']
    
    return int(c1 or c2 or c3)


@experiment.hook('add_graph_metadata', default=False, replace=True)
def add_graph_metadata(e: Experiment, data: dict, graph: dict) -> dict:
    """
    We add the compound id for identification and the molecular weight
    """
    pass #graph['graph_id'] = data['ID']


@experiment.hook('load_dataset', default=False, replace=True)
def load_dataset(e: Experiment) -> Dict[int, dict]:
    df = pd.read_csv('/media/data/Downloads/zinc_250k.csv')
    e.log(f'dataset of size {len(df)} loaded')
    e.log(df.head())

    dataset: Dict[int, dict] = {}
    index: int = 0
    for c, data in enumerate(df.to_dict('records')):
        data['smiles'] = data['smiles']
        
        # == MOLECULE FILTERS ==
        # We don't want to use compounds with '.' in the smiles (separate molecules)
        if '.' in data['smiles']:
            continue
        
        # We don't want to use compounds that only consist of a single atom
        mol = Chem.MolFromSmiles(data['smiles'])
        if not mol:
            continue
        
        # We also don't want to accept "molecules" that are essentially just individual atoms
        if len(mol.GetAtoms()) < 2:
            continue
        
        # == TARGETS ==
        
        # Use the previously defined function to calculate the binary target label based on the 
        # local smarts patterns.
        label: int = motif_label(mol, SMARTS_PATTERNS_COMPILED)
        
        # In the end we dont need the integer but we need a one-hot encoded vector to represent the 
        # target classification label, which we construct here.
        graph_labels = np.array([1 - label, label], dtype=np.float32)
        
        data['targets'] = graph_labels
        dataset[index] = data
        
        index += 1
        
        if c % 1000 == 0:
            print(f' * {c} elements loaded')

    e.log('returning dataset...')
    return dataset

experiment.run_if_main()


import rdkit.Chem as Chem
from typing import List, Dict
from pycomex.functional.experiment import Experiment
from pycomex.utils import folder_path, file_namespace

from chem_mat_data import load_smiles_dataset
DATASET_NAME: str = 'bbbp'

__TESTING__ = False

# This is the dataset template.
experiment = Experiment.extend(
    'create_graph_datasets.py',
    base_path=folder_path(__file__),
    namespace=file_namespace(__file__),
    glob=globals(),
)

@experiment.hook('add_graph_metadata', default=False, replace=True)
def add_graph_metadata(e: Experiment, data: dict, graph: dict) -> dict:
    """
    We add the names of the compounds
    """
    graph['graph_name'] = data['name']


@experiment.hook('load_dataset', default=False, replace=True)
def load_dataset(e: Experiment) -> Dict[int, dict]:
    df = load_smiles_dataset('BBBP')

    dataset: Dict[int, dict] = {}
    for index, data in enumerate(df.to_dict('records')):
        data['smiles'] = data['smiles']
        
        # == MOLECULE FILTERS ==
        # We don't want to use compounds with '.' in the smiles (separate molecules)
        if '.' in data['smiles']:
            continue
        
        # We don't want to use compounds that only consist of a single atom
        mol = Chem.MolFromSmiles(data['smiles'])
        if not mol:
            continue
        
        # We also don't want to accept "molecules" that are essentially just individual atoms
        if len(mol.GetAtoms()) < 2:
            continue
        
        #p_np is either 0 or 1
        data['targets'] = [data['p_np'] == index for index in range(2)]
        dataset[index] = data
        
    return dataset

experiment.run_if_main()




import rdkit.Chem as Chem
from typing import List, Dict
from pycomex.functional.experiment import Experiment
from pycomex.utils import folder_path, file_namespace

from chem_mat_data import load_smiles_dataset

DATASET_NAME: str = 'freesolv'
DESCRIPTION: str = (
    'A collection of experimental and calculated hydration free energies for small molecules in water. '
    'The calculated values are derived from alchemical free energy calculations using molecular dynamics simulations. '
    'Each molecule is associated with 2 target values (experimental, calculated) where the first '
    'value is the experimental hydration free energy and the second value is the calculated hydration free energy.'
)

# :param METADATA:
#       A dictionary which will be used as the basis for the metadata that will be added 
#       as additional information to the file share server.
METADATA: dict = {
    'tags': [
        'Molecules', 
        'SMILES', 
        'Experimental',
        'Hydration Free Energy',
        'Molecular Dynamics',
        'Hydration'
    ],
    'verbose': 'Hydration Free Energy',
    'sources': [
        'https://pubmed.ncbi.nlm.nih.gov/24928188/',
        'https://pubs.acs.org/doi/10.1021/acs.jcim.8b00180',
        'https://github.com/MobleyLab/FreeSolv',
    ],
    # TADF: Thermally Activated Delayed Fluorescence
    'target_descriptions': {
        '0': 'HDF expt - Experimentally determined hydration free energy',
        '1': 'HDF calc - hydration free energy calculated using molecular dynamics simulations',
    }
}

experiment = Experiment.extend(
    'create_graph_datasets.py',
    base_path=folder_path(__file__),
    namespace=file_namespace(__file__),
    glob=globals(),
)

@experiment.hook('add_graph_metadata', default=False, replace=True)
def add_graph_metadata(e: Experiment, data: dict, graph: dict) -> dict:
    """
    In this experiment we add the iupac nomenclature as additional metadata field for possible identification
    """
    graph['graph_id'] = data['iupac']

@experiment.hook('load_dataset', default=False, replace=True)
def load_dataset(e: Experiment) -> Dict[int, dict]:
    
    df = load_smiles_dataset('freesolv')
    dataset: Dict[int, dict] = {}
    for index, data in enumerate(df.to_dict('records')):

        data['smiles'] = data['smiles']
        
        # == MOLECULE FILTERS ==
        # We don't want to use compounds with '.' in the smiles (separate molecules)
        if '.' in data['smiles']:
            continue
        
        # We don't want to use compounds that only consist of a single atom
        mol = Chem.MolFromSmiles(data['smiles'])
        if not mol:
            continue
        
        # We also don't want to accept "molecules" that are essentially just individual atoms
        if len(mol.GetAtoms()) < 2:
            continue
        
        # == TARGET VALUES ==
        # Only one target. We combine the experimental and computational data into a single list 
        data['targets'] = [data['expt'], data['calc']]

        dataset[index] = data

    return dataset

experiment.run_if_main()



"""Nox configuration for chem_mat_data package."""

import nox

# Python versions to test against
PYTHON_VERSIONS = ["3.8", "3.9", "3.10", "3.11", "3.12"]


@nox.session(python=PYTHON_VERSIONS[0], venv_backend="uv")
def test(session):
    """Run unit tests with pytest."""
    session.install(".[dev]")
    # Skip tests that require network access only
    session.run(
        "pytest",
        "tests/",
        "-v",
        "--ignore=tests/test_datasets.py",     # Network timeouts
        "-k", "not localonly"  # Skip local-only tests
    )

@nox.session(python="3.11", venv_backend="uv")
def lint(session):
    """Run linting with ruff."""
    session.install(".[dev]")
    session.run("ruff", "check", ".")


@nox.session(python="3.11", venv_backend="uv")
def format(session):
    """Format code with ruff."""
    session.install(".[dev]")
    session.run("ruff", "format", ".")


@nox.session(python="3.11", venv_backend="uv")
def type_check(session):
    """Run type checking with mypy."""
    session.install(".[dev]")
    session.run("mypy", "chem_mat_data", "--ignore-missing-imports")


@nox.session(python="3.11", venv_backend="uv")
def docs(session):
    """Build documentation."""
    session.install(".[dev]")
    session.run("sphinx-build", "-b", "html", "docs", "docs/_build/html")


@nox.session(python=PYTHON_VERSIONS, venv_backend="uv")
def install_check(session):
    """Check that the package can be installed and imported."""
    session.install(".")
    session.run("python", "-c", "import chem_mat_data; print('Import successful')")


@nox.session(python="3.11", venv_backend="uv")
def build(session):
    """Build the package."""
    session.install(".[dev]")
    session.run("python", "-m", "build")


@nox.session(python="3.11", venv_backend="uv")
def clean(session):
    """Clean build artifacts."""
    import shutil
    import os

    # Directories to clean
    clean_dirs = [
        "build",
        "dist",
        "*.egg-info",
        ".pytest_cache",
        ".coverage",
        "htmlcov",
        ".mypy_cache",
        ".ruff_cache"
    ]

    for pattern in clean_dirs:
        if "*" in pattern:
            import glob
            for path in glob.glob(pattern):
                if os.path.isdir(path):
                    shutil.rmtree(path)
                elif os.path.isfile(path):
                    os.remove(path)
        else:
            if os.path.isdir(pattern):
                shutil.rmtree(pattern)
            elif os.path.isfile(pattern):
                os.remove(pattern)

    session.log("Cleaned build artifacts")


#!/usr/bin/env python3
"""
Test script to download and parse a larger HOPV15 dataset file to debug the parser.
"""

import requests
import tempfile
import os
from chem_mat_data.data import HOPV15Parser

def download_hopv15_dataset():
    """Download the HOPV15 dataset from the official source."""
    url = 'https://figshare.com/ndownloader/files/4513735'

    print("Downloading HOPV15 dataset...")
    try:
        # Download with SSL verification disabled as in the original script
        response = requests.get(url, verify=False, stream=True, timeout=600)  # 10 minute timeout
        response.raise_for_status()

        # Save to temporary file
        with tempfile.NamedTemporaryFile(mode='wb', delete=False, suffix='.data') as f:
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:
                    f.write(chunk)
            temp_path = f.name

        print(f"Downloaded dataset to: {temp_path}")
        file_size = os.path.getsize(temp_path)
        print(f"File size: {file_size:,} bytes")

        return temp_path

    except Exception as e:
        print(f"Error downloading dataset: {e}")
        return None

def test_hopv15_parser(file_path):
    """Test the HOPV15 parser on the downloaded dataset."""
    print(f"\nTesting HOPV15Parser on: {file_path}")

    try:
        # Initialize parser
        parser = HOPV15Parser(path=file_path)
        print("✓ Parser initialized successfully")

        # Test single parse (first molecule)
        print("\nTesting single parse...")
        try:
            mol, info = parser.parse()
            print("✓ Single parse successful")
            print(f"  SMILES: {info.get('smiles', 'None')}")
            print(f"  Molecule atoms: {mol.GetNumAtoms() if mol else 'None'}")
            print(f"  Has experimental properties: {'experimental_properties' in info}")
            print(f"  Number of conformers: {len(info.get('conformers', []))}")
        except Exception as e:
            print(f"✗ Single parse failed: {e}")
            import traceback
            traceback.print_exc()
            return False

        # Test parse_all (all molecules)
        print("\nTesting parse_all...")
        try:
            molecules = parser.parse_all()
            print(f"✓ Parse_all successful - found {len(molecules)} molecules")

            # Show statistics
            with_exp_props = sum(1 for _, info in molecules if 'experimental_properties' in info)
            total_conformers = sum(len(info.get('conformers', [])) for _, info in molecules)

            print(f"  Molecules with experimental properties: {with_exp_props}")
            print(f"  Total conformers across all molecules: {total_conformers}")

            # Show first few molecules
            print(f"\nFirst 3 molecules:")
            for i, (mol, info) in enumerate(molecules[:3]):
                print(f"  {i+1}. {info['smiles']} ({mol.GetNumAtoms()} atoms)")
                if 'experimental_properties' in info:
                    props = info['experimental_properties']
                    print(f"     PCE: {props.get('PCE', 'N/A')}, HOMO: {props.get('HOMO', 'N/A')}")

        except Exception as e:
            print(f"✗ Parse_all failed: {e}")
            import traceback
            traceback.print_exc()
            return False

        return True

    except Exception as e:
        print(f"✗ Parser initialization failed: {e}")
        import traceback
        traceback.print_exc()
        return False

def main():
    """Main test function."""
    print("HOPV15 Parser Test on Large Dataset")
    print("=" * 50)

    # Download dataset
    dataset_path = download_hopv15_dataset()
    if not dataset_path:
        print("Failed to download dataset, exiting...")
        return

    try:
        # Test parser
        success = test_hopv15_parser(dataset_path)

        if success:
            print("\n✓ All tests passed successfully!")
        else:
            print("\n✗ Some tests failed!")

    finally:
        # Clean up
        if dataset_path and os.path.exists(dataset_path):
            os.unlink(dataset_path)
            print(f"\nCleaned up temporary file: {dataset_path}")

if __name__ == "__main__":
    main()


import os
from typing import List, Dict
import pandas as pd
import gzip
import shutil
from rdkit import Chem
from pycomex.functional.experiment import Experiment
from pycomex.utils import folder_path, file_namespace

from chem_mat_data.config import Config
from chem_mat_data.web import NextcloudFileShare
from chem_mat_data.main import get_file_share

# :param DATASET_NAME:
#       This is the name of the dataset that will be used to identify the dataset in the
#       file share server. It will also be used to create the folder structure for the dataset
#       on the file share server.
DATASET_NAME: str = 'skin_irritation'
# :param SMILES_COLUMN:
#       This is the string name of the CSV column which contains the SMILES strings of
#       the molecules.
SMILES_COLUMN: str = 'SMILES'
# :param TARGET_COLUMNS:
#       This is a list of string names of the CSV columns which contain the target values
#       of the dataset. This can be a single column for regression tasks or multiple columns
#       for multi-target regression or classification tasks. For the final graph dataset
#       the target values will be merged into a single numeric vector that contains the 
#       corresponding values in the same order as the column names are defined here.
TARGET_COLUMNS: List[str] = ['label']
# :param DATASET_TYPE:
#       Either 'regression' or 'classification' to define the type of the dataset. This
#       will also determine how the target values are processed.
DATASET_TYPE: str = 'classification'
# :param DESCRIPTION:
#       This is a string description of the dataset that will be stored in the experiment
#       metadata.
DESCRIPTION: str = (
    'Curated dataset of roughly 1000 molecules focused on the skin irritation potential of small organic molecules.'
    'Data were retrieved from the Registration, Evaluation, Authorization and Restriction of Chemicals (REACH) database, and were curated as part of the study by Borba et al.'    
)
# :param METADATA:
#       A dictionary which will be used as the basis for the metadata that will be added 
#       as additional information to the file share server.
METADATA: dict = {
    'tags': [
        'Molecules', 
        'SMILES', 
        'Biology', 
    ],
    'sources': [
        'https://db.chempharos.eu/datasets/Datasets.zul?datasetID=ds14',
        'https://ehp.niehs.nih.gov/doi/10.1289/EHP9341',    
    ],
    'target_descriptions': {
        '0': 'non-irritant',
        '1': 'skin-irritant',
    }
}

__TESTING__ = False

experiment = Experiment.extend(
    'create_graph_datasets.py',
    base_path=folder_path(__file__),
    namespace=file_namespace(__file__),
    glob=globals(),
)

@experiment.hook('add_graph_metadata', default=False, replace=True)
def add_graph_metadata(e: Experiment, data: dict, graph: dict) -> dict:
    """
    We add the compound id for identification and the molecular weight
    """
    #graph['graph_name'] = data['Name']
    graph['graph_subset'] = data['dataset']


@experiment.hook('load_dataset', default=False, replace=True)
def load_dataset(e: Experiment) -> Dict[int, dict]:
    
    ## -- Load Dataset --
    e.log('Loading the EXCEL file from the remote file share server...')
    config = Config()
    file_share: NextcloudFileShare = get_file_share(config)
    file_path: str = file_share.download_file('skin_irritation_dataset.xlsx', folder_path=e.path)
    df: pd.DataFrame = pd.read_excel(file_path)
    print(df.head())
    
    ## -- Save Dataset --
    e.log('Saving the dataset as CSV and GZipped CSV file...')
    csv_path = os.path.join(e.path, f'{e.DATASET_NAME}.csv')
    df.to_csv(csv_path, index=False)

    gz_path = csv_path + '.gz'
    with open(csv_path, 'rb') as f_in, gzip.open(gz_path, 'wb') as f_out:
        shutil.copyfileobj(f_in, f_out)

    ## -- Processing Dataset --
    dataset: Dict[int, dict] = {}
    index: int = 0
    for data in df.to_dict('records'):
        
        data['smiles'] = data[e.SMILES_COLUMN]
        
        ## -- Molecule Filters --
        # We don't want to use compounds with '.' in the smiles (separate molecules)
        if '.' in data['smiles']:
            continue
        
        # We don't want to use compounds that only consist of a single atom
        mol = Chem.MolFromSmiles(data['smiles'])
        if not mol:
            continue
        
        # We also don't want to accept "molecules" that are essentially just individual atoms
        if len(mol.GetAtoms()) < 2:
            continue
        
        ## -- Target Values --
        # In this dataset we only have one target
        data['targets'] = [0, 1] if data['Label'] else [1, 0]
        dataset[index] = data
        
        index += 1

    return dataset

experiment.run_if_main()
